# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2022 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import base64
import json
import logging
import typing as t

from . import errors
from .credentials.errors import CredentialsNotFound, StoreError
from .credentials.store import Store
from .utils import config

if t.TYPE_CHECKING:
    from ansible_collections.community.docker.plugins.module_utils._api.api.client import (
        APIClient,
    )


INDEX_NAME = "docker.io"
INDEX_URL = f"https://index.{INDEX_NAME}/v1/"
TOKEN_USERNAME = "<token>"

log = logging.getLogger(__name__)


def resolve_repository_name(repo_name: str) -> tuple[str, str]:
    if "://" in repo_name:
        raise errors.InvalidRepository(
            f"Repository name cannot contain a scheme ({repo_name})"
        )

    index_name, remote_name = split_repo_name(repo_name)
    if index_name[0] == "-" or index_name[-1] == "-":
        raise errors.InvalidRepository(
            f"Invalid index name ({index_name}). Cannot begin or end with a hyphen."
        )
    return resolve_index_name(index_name), remote_name


def resolve_index_name(index_name: str) -> str:
    index_name = convert_to_hostname(index_name)
    if index_name == "index." + INDEX_NAME:
        index_name = INDEX_NAME
    return index_name


def get_config_header(client: APIClient, registry: str) -> bytes | None:
    log.debug("Looking for auth config")
    if not client._auth_configs or client._auth_configs.is_empty:
        log.debug("No auth config in memory - loading from filesystem")
        client._auth_configs = load_config(credstore_env=client.credstore_env)
    authcfg = resolve_authconfig(
        client._auth_configs, registry, credstore_env=client.credstore_env
    )
    # Do not fail here if no authentication exists for this
    # specific registry as we can have a readonly pull. Just
    # put the header if we can.
    if authcfg:
        log.debug("Found auth config")
        # auth_config needs to be a dict in the format used by
        # auth.py username , password, serveraddress, email
        return encode_header(authcfg)
    log.debug("No auth config found")
    return None


def split_repo_name(repo_name: str) -> tuple[str, str]:
    parts = repo_name.split("/", 1)
    if len(parts) == 1 or (
        "." not in parts[0] and ":" not in parts[0] and parts[0] != "localhost"
    ):
        # This is a docker index repo (ex: username/foobar or ubuntu)
        return INDEX_NAME, repo_name
    return tuple(parts)  # type: ignore


def get_credential_store(
    authconfig: dict[str, t.Any] | AuthConfig, registry: str
) -> str | None:
    if not isinstance(authconfig, AuthConfig):
        authconfig = AuthConfig(authconfig)
    return authconfig.get_credential_store(registry)


class AuthConfig(dict):
    def __init__(
        self, dct: dict[str, t.Any], credstore_env: dict[str, str] | None = None
    ):
        if "auths" not in dct:
            dct["auths"] = {}
        self.update(dct)
        self._credstore_env = credstore_env
        self._stores: dict[str, Store] = {}

    @classmethod
    def parse_auth(
        cls, entries: dict[str, dict[str, t.Any]], raise_on_error: bool = False
    ) -> dict[str, dict[str, t.Any]]:
        """
        Parses authentication entries

        Args:
          entries:        Dict of authentication entries.
          raise_on_error: If set to true, an invalid format will raise
                          InvalidConfigFile

        Returns:
          Authentication registry.
        """

        conf: dict[str, dict[str, t.Any]] = {}
        for registry, entry in entries.items():
            if not isinstance(entry, dict):
                log.debug("Config entry for key %s is not auth config", registry)  # type: ignore
                # We sometimes fall back to parsing the whole config as if it
                # was the auth config by itself, for legacy purposes. In that
                # case, we fail silently and return an empty conf if any of the
                # keys is not formatted properly.
                if raise_on_error:
                    raise errors.InvalidConfigFile(
                        f"Invalid configuration for registry {registry}"
                    )
                return {}
            if "identitytoken" in entry:
                log.debug("Found an IdentityToken entry for registry %s", registry)
                conf[registry] = {"IdentityToken": entry["identitytoken"]}
                continue  # Other values are irrelevant if we have a token

            if "auth" not in entry:
                # Starting with engine v1.11 (API 1.23), an empty dictionary is
                # a valid value in the auths config.
                # https://github.com/docker/compose/issues/3265
                log.debug(
                    "Auth data for %s is absent. Client might be using a credentials store instead.",
                    registry,
                )
                conf[registry] = {}
                continue

            username, password = decode_auth(entry["auth"])
            log.debug(
                "Found entry (registry=%s, username=%s)", repr(registry), repr(username)
            )

            conf[registry] = {
                "username": username,
                "password": password,
                "email": entry.get("email"),
                "serveraddress": registry,
            }
        return conf

    @classmethod
    def load_config(
        cls,
        config_path: str | None,
        config_dict: dict[str, t.Any] | None,
        credstore_env: dict[str, str] | None = None,
    ) -> t.Self:
        """
        Loads authentication data from a Docker configuration file in the given
        root directory or if config_path is passed use given path.
        Lookup priority:
            explicit config_path parameter > DOCKER_CONFIG environment
            variable > ~/.docker/config.json > ~/.dockercfg
        """

        if not config_dict:
            config_file = config.find_config_file(config_path)

            if not config_file:
                return cls({}, credstore_env)
            try:
                with open(config_file, "rt", encoding="utf-8") as f:
                    config_dict = json.load(f)
            except (IOError, KeyError, ValueError) as e:
                # Likely missing new Docker config file or it is in an
                # unknown format, continue to attempt to read old location
                # and format.
                log.debug(e)
                return cls(_load_legacy_config(config_file), credstore_env)

        res = {}
        if config_dict.get("auths"):
            log.debug("Found 'auths' section")
            res.update(
                {"auths": cls.parse_auth(config_dict.pop("auths"), raise_on_error=True)}
            )
        if config_dict.get("credsStore"):
            log.debug("Found 'credsStore' section")
            res.update({"credsStore": config_dict.pop("credsStore")})
        if config_dict.get("credHelpers"):
            log.debug("Found 'credHelpers' section")
            res.update({"credHelpers": config_dict.pop("credHelpers")})
        if res:
            return cls(res, credstore_env)

        log.debug(
            "Could not find auth-related section ; attempting to interpret "
            "as auth-only file"
        )
        return cls({"auths": cls.parse_auth(config_dict)}, credstore_env)

    @property
    def auths(self) -> dict[str, dict[str, t.Any]]:
        return self.get("auths", {})

    @property
    def creds_store(self) -> str | None:
        return self.get("credsStore", None)

    @property
    def cred_helpers(self) -> dict[str, t.Any]:
        return self.get("credHelpers", {})

    @property
    def is_empty(self) -> bool:
        return not self.auths and not self.creds_store and not self.cred_helpers

    def resolve_authconfig(
        self, registry: str | None = None
    ) -> dict[str, t.Any] | None:
        """
        Returns the authentication data from the given auth configuration for a
        specific registry. As with the Docker client, legacy entries in the
        config with full URLs are stripped down to hostnames before checking
        for a match. Returns None if no match was found.
        """

        if self.creds_store or self.cred_helpers:
            store_name = self.get_credential_store(registry)
            if store_name is not None:
                log.debug('Using credentials store "%s"', store_name)
                cfg = self._resolve_authconfig_credstore(registry, store_name)
                if cfg is not None:
                    return cfg
                log.debug("No entry in credstore - fetching from auth dict")

        # Default to the public index server
        registry = resolve_index_name(registry) if registry else INDEX_NAME
        log.debug("Looking for auth entry for %s", repr(registry))

        if registry in self.auths:
            log.debug("Found %s", repr(registry))
            return self.auths[registry]

        for key, conf in self.auths.items():
            if resolve_index_name(key) == registry:
                log.debug("Found %s", repr(key))
                return conf

        log.debug("No entry found")
        return None

    def _resolve_authconfig_credstore(
        self, registry: str | None, credstore_name: str
    ) -> dict[str, t.Any] | None:
        if not registry or registry == INDEX_NAME:
            # The ecosystem is a little schizophrenic with index.docker.io VS
            # docker.io - in that case, it seems the full URL is necessary.
            registry = INDEX_URL
        log.debug("Looking for auth entry for %s", repr(registry))
        store = self._get_store_instance(credstore_name)
        try:
            data = store.get(registry)
            res = {
                "ServerAddress": registry,
            }
            if data["Username"] == TOKEN_USERNAME:
                res["IdentityToken"] = data["Secret"]
            else:
                res.update(
                    {
                        "Username": data["Username"],
                        "Password": data["Secret"],
                    }
                )
            return res
        except CredentialsNotFound:
            log.debug("No entry found")
            return None
        except StoreError as e:
            raise errors.DockerException(f"Credentials store error: {e}") from e

    def _get_store_instance(self, name: str) -> Store:
        if name not in self._stores:
            self._stores[name] = Store(name, environment=self._credstore_env)
        return self._stores[name]

    def get_credential_store(self, registry: str | None) -> str | None:
        if not registry or registry == INDEX_NAME:
            registry = INDEX_URL

        return self.cred_helpers.get(registry) or self.creds_store

    def get_all_credentials(self) -> dict[str, dict[str, t.Any] | None]:
        auth_data: dict[str, dict[str, t.Any] | None] = self.auths.copy()  # type: ignore
        if self.creds_store:
            # Retrieve all credentials from the default store
            store = self._get_store_instance(self.creds_store)
            for k in store.list():
                auth_data[k] = self._resolve_authconfig_credstore(k, self.creds_store)
                auth_data[convert_to_hostname(k)] = auth_data[k]

        # credHelpers entries take priority over all others
        for reg, store_name in self.cred_helpers.items():
            auth_data[reg] = self._resolve_authconfig_credstore(reg, store_name)
            auth_data[convert_to_hostname(reg)] = auth_data[reg]

        return auth_data

    def add_auth(self, reg: str, data: dict[str, t.Any]) -> None:
        self["auths"][reg] = data


def resolve_authconfig(
    authconfig: AuthConfig | dict[str, t.Any],
    registry: str | None = None,
    credstore_env: dict[str, str] | None = None,
) -> dict[str, t.Any] | None:
    if not isinstance(authconfig, AuthConfig):
        authconfig = AuthConfig(authconfig, credstore_env)
    return authconfig.resolve_authconfig(registry)


def convert_to_hostname(url: str) -> str:
    return url.replace("http://", "").replace("https://", "").split("/", 1)[0]


def decode_auth(auth: str | bytes) -> tuple[str, str]:
    if isinstance(auth, str):
        auth = auth.encode("ascii")
    s = base64.b64decode(auth)
    login, pwd = s.split(b":", 1)
    return login.decode("utf8"), pwd.decode("utf8")


def encode_header(auth: dict[str, t.Any]) -> bytes:
    auth_json = json.dumps(auth).encode("ascii")
    return base64.urlsafe_b64encode(auth_json)


def parse_auth(
    entries: dict[str, dict[str, t.Any]], raise_on_error: bool = False
) -> dict[str, dict[str, t.Any]]:
    """
    Parses authentication entries

    Args:
      entries:        Dict of authentication entries.
      raise_on_error: If set to true, an invalid format will raise
                      InvalidConfigFile

    Returns:
      Authentication registry.
    """

    return AuthConfig.parse_auth(entries, raise_on_error)


def load_config(
    config_path: str | None = None,
    config_dict: dict[str, t.Any] | None = None,
    credstore_env: dict[str, str] | None = None,
) -> AuthConfig:
    return AuthConfig.load_config(config_path, config_dict, credstore_env)


def _load_legacy_config(config_file: str) -> dict[str, dict[str, t.Any]]:
    log.debug("Attempting to parse legacy auth file format")
    try:
        data = []
        with open(config_file, "rt", encoding="utf-8") as f:
            for line in f.readlines():
                data.append(line.strip().split(" = ")[1])
            if len(data) < 2:
                # Not enough data
                raise errors.InvalidConfigFile("Invalid or empty configuration file!")

        username, password = decode_auth(data[0])
        return {
            "auths": {
                INDEX_NAME: {
                    "username": username,
                    "password": password,
                    "email": data[1],
                    "serveraddress": INDEX_URL,
                }
            }
        }
    except Exception as e:  # pylint: disable=broad-exception-caught
        log.debug(e)

    log.debug("All parsing attempts failed - returning empty config")
    return {}



#!/usr/bin/python

# Copyright (c) 2020, Christian Wollinger <cwollinger@web.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipwcli_dns

short_description: Manage DNS Records for Ericsson IPWorks using C(ipwcli)

version_added: '0.2.0'

description:
  - Manage DNS records for the Ericsson IPWorks DNS server. The module uses the C(ipwcli) to deploy the DNS records.
requirements:
  - ipwcli (installed on Ericsson IPWorks)

notes:
  - To make the DNS record changes effective, you need to run C(update dnsserver) on the ipwcli.
extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  dnsname:
    description:
      - Name of the record.
    required: true
    type: str
  type:
    description:
      - Type of the record.
    required: true
    type: str
    choices: [NAPTR, SRV, A, AAAA]
  container:
    description:
      - Sets the container zone for the record.
    required: true
    type: str
  address:
    description:
      - The IP address for the A or AAAA record.
      - Required for O(type=A) or O(type=AAAA).
    type: str
  ttl:
    description:
      - Sets the TTL of the record.
    type: int
    default: 3600
  state:
    description:
      - Whether the record should exist or not.
    type: str
    choices: [absent, present]
    default: present
  priority:
    description:
      - Sets the priority of the SRV record.
    type: int
    default: 10
  weight:
    description:
      - Sets the weight of the SRV record.
    type: int
    default: 10
  port:
    description:
      - Sets the port of the SRV record.
      - Required for O(type=SRV).
    type: int
  target:
    description:
      - Sets the target of the SRV record.
      - Required for O(type=SRV).
    type: str
  order:
    description:
      - Sets the order of the NAPTR record.
      - Required for O(type=NAPTR).
    type: int
  preference:
    description:
      - Sets the preference of the NAPTR record.
      - Required for O(type=NAPTR).
    type: int
  flags:
    description:
      - Sets one of the possible flags of NAPTR record.
      - Required for O(type=NAPTR).
    type: str
    choices: ['S', 'A', 'U', 'P']
  service:
    description:
      - Sets the service of the NAPTR record.
      - Required for O(type=NAPTR).
    type: str
  replacement:
    description:
      - Sets the replacement of the NAPTR record.
      - Required for O(type=NAPTR).
    type: str
  username:
    description:
      - Username to login on ipwcli.
    type: str
    required: true
  password:
    description:
      - Password to login on ipwcli.
    type: str
    required: true

author:
  - Christian Wollinger (@cwollinger)
"""

EXAMPLES = r"""
- name: Create A record
  community.general.ipwcli_dns:
    dnsname: example.com
    type: A
    container: ZoneOne
    address: 127.0.0.1

- name: Remove SRV record if exists
  community.general.ipwcli_dns:
    dnsname: _sip._tcp.test.example.com
    type: SRV
    container: ZoneOne
    ttl: 100
    state: absent
    target: example.com
    port: 5060

- name: Create NAPTR record
  community.general.ipwcli_dns:
    dnsname: test.example.com
    type: NAPTR
    preference: 10
    container: ZoneOne
    ttl: 100
    order: 10
    service: 'SIP+D2T'
    replacement: '_sip._tcp.test.example.com.'
    flags: S
"""

RETURN = r"""
record:
  description: The created record from the input params.
  type: str
  returned: always
"""

from ansible.module_utils.basic import AnsibleModule


class ResourceRecord:
    def __init__(self, module):
        self.module = module
        self.dnsname = module.params["dnsname"]
        self.dnstype = module.params["type"]
        self.container = module.params["container"]
        self.address = module.params["address"]
        self.ttl = module.params["ttl"]
        self.state = module.params["state"]
        self.priority = module.params["priority"]
        self.weight = module.params["weight"]
        self.port = module.params["port"]
        self.target = module.params["target"]
        self.order = module.params["order"]
        self.preference = module.params["preference"]
        self.flags = module.params["flags"]
        self.service = module.params["service"]
        self.replacement = module.params["replacement"]
        self.user = module.params["username"]
        self.password = module.params["password"]

    def create_naptrrecord(self):
        # create NAPTR record with the given params
        record = (
            f"naptrrecord {self.dnsname} -set ttl={self.ttl};container={self.container};order={self.order};"
            f'preference={self.preference};flags="{self.flags}";service="{self.service}";replacement="{self.replacement}"'
        )
        return record

    def create_srvrecord(self):
        # create SRV record with the given params
        record = (
            f"srvrecord {self.dnsname} -set ttl={self.ttl};container={self.container};priority={self.priority};"
            f"weight={self.weight};port={self.port};target={self.target}"
        )
        return record

    def create_arecord(self):
        # create A record with the given params
        if self.dnstype == "AAAA":
            record = f"aaaarecord {self.dnsname} {self.address} -set ttl={self.ttl};container={self.container}"
        else:
            record = f"arecord {self.dnsname} {self.address} -set ttl={self.ttl};container={self.container}"

        return record

    def list_record(self, record):
        # check if the record exists via list on ipwcli
        search = f"list {record.replace(';', '&&').replace('set', 'where')}"
        cmd = [
            self.module.get_bin_path("ipwcli", True),
            f"-user={self.user}",
            f"-password={self.password}",
        ]
        rc, out, err = self.module.run_command(cmd, data=search)

        if "Invalid username or password" in out:
            self.module.fail_json(msg="access denied at ipwcli login: Invalid username or password")

        if (
            (f"ARecord {self.dnsname}" in out and rc == 0)
            or (f"SRVRecord {self.dnsname}" in out and rc == 0)
            or (f"NAPTRRecord {self.dnsname}" in out and rc == 0)
        ):
            return True, rc, out, err

        return False, rc, out, err

    def deploy_record(self, record):
        # check what happens if create fails on ipworks
        stdin = f"create {record}"
        cmd = [
            self.module.get_bin_path("ipwcli", True),
            f"-user={self.user}",
            f"-password={self.password}",
        ]
        rc, out, err = self.module.run_command(cmd, data=stdin)

        if "Invalid username or password" in out:
            self.module.fail_json(msg="access denied at ipwcli login: Invalid username or password")

        if "1 object(s) created." in out:
            return rc, out, err
        else:
            self.module.fail_json(msg="record creation failed", stderr=out)

    def delete_record(self, record):
        # check what happens if create fails on ipworks
        stdin = f"delete {record.replace(';', '&&').replace('set', 'where')}"
        cmd = [
            self.module.get_bin_path("ipwcli", True),
            f"-user={self.user}",
            f"-password={self.password}",
        ]
        rc, out, err = self.module.run_command(cmd, data=stdin)

        if "Invalid username or password" in out:
            self.module.fail_json(msg="access denied at ipwcli login: Invalid username or password")

        if "1 object(s) were updated." in out:
            return rc, out, err
        else:
            self.module.fail_json(msg="record deletion failed", stderr=out)


def run_module():
    # define available arguments/parameters a user can pass to the module
    module_args = dict(
        dnsname=dict(type="str", required=True),
        type=dict(type="str", required=True, choices=["A", "AAAA", "SRV", "NAPTR"]),
        container=dict(type="str", required=True),
        address=dict(type="str"),
        ttl=dict(type="int", default=3600),
        state=dict(type="str", default="present", choices=["absent", "present"]),
        priority=dict(type="int", default=10),
        weight=dict(type="int", default=10),
        port=dict(type="int"),
        target=dict(type="str"),
        order=dict(type="int"),
        preference=dict(type="int"),
        flags=dict(type="str", choices=["S", "A", "U", "P"]),
        service=dict(type="str"),
        replacement=dict(type="str"),
        username=dict(type="str", required=True),
        password=dict(type="str", required=True, no_log=True),
    )

    # define result
    result = dict(changed=False, stdout="", stderr="", rc=0, record="")

    # supports check mode
    module = AnsibleModule(
        argument_spec=module_args,
        required_if=[
            ["type", "A", ["address"]],
            ["type", "AAAA", ["address"]],
            ["type", "SRV", ["port", "target"]],
            ["type", "NAPTR", ["preference", "order", "service", "replacement"]],
        ],
        supports_check_mode=True,
    )

    user = ResourceRecord(module)

    if user.dnstype == "NAPTR":
        record = user.create_naptrrecord()
    elif user.dnstype == "SRV":
        record = user.create_srvrecord()
    elif user.dnstype == "A" or user.dnstype == "AAAA":
        record = user.create_arecord()

    found, rc, out, err = user.list_record(record)

    if found and user.state == "absent":
        if module.check_mode:
            module.exit_json(changed=True)
        rc, out, err = user.delete_record(record)
        result["changed"] = True
        result["record"] = record
        result["rc"] = rc
        result["stdout"] = out
        result["stderr"] = err
    elif not found and user.state == "present":
        if module.check_mode:
            module.exit_json(changed=True)
        rc, out, err = user.deploy_record(record)
        result["changed"] = True
        result["record"] = record
        result["rc"] = rc
        result["stdout"] = out
        result["stderr"] = err
    else:
        result["changed"] = False
        result["record"] = record
        result["rc"] = rc
        result["stdout"] = out
        result["stderr"] = err

    module.exit_json(**result)


def main():
    run_module()


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import re
import typing as t
from dataclasses import dataclass

_PATH_RE = re.compile(
    r"^[a-z0-9]+((\.|_|__|-+)[a-z0-9]+)*(\/[a-z0-9]+((\.|_|__|-+)[a-z0-9]+)*)*$"
)
_TAG_RE = re.compile(r"^[a-zA-Z0-9_][a-zA-Z0-9._-]{0,127}$")
_DIGEST_RE = re.compile(r"^sha256:[0-9a-fA-F]{64}$")


def is_digest(name: str, allow_empty: bool = False) -> bool:
    """Check whether the given name is in fact an image ID (hash)."""
    if not name:
        return allow_empty
    return _DIGEST_RE.match(name) is not None


def is_tag(name: str, allow_empty: bool = False) -> bool:
    """Check whether the given name can be an image tag."""
    if not name:
        return allow_empty
    return _TAG_RE.match(name) is not None


@dataclass
class ImageName:
    registry: str | None
    path: str
    tag: str | None
    digest: str | None

    @classmethod
    def parse(cls, name: str) -> t.Self:
        registry: str | None = None
        tag: str | None = None
        digest: str | None = None
        parts = name.rsplit("@", 1)
        if len(parts) == 2:
            name, digest = parts
        parts = name.rsplit(":", 1)
        if len(parts) == 2 and "/" not in parts[1]:
            name, tag = parts
        parts = name.split("/", 1)
        if len(parts) == 2 and (
            "." in parts[0] or ":" in parts[0] or parts[0] == "localhost"
        ):
            registry, name = parts
        return cls(registry, name, tag, digest)

    def validate(self) -> t.Self:
        if self.registry:
            if self.registry[0] == "-" or self.registry[-1] == "-":
                raise ValueError(
                    f'Invalid registry name ({self.registry}): must not begin or end with a "-".'
                )
            if self.registry[-1] == ":":
                raise ValueError(
                    f'Invalid registry name ({self.registry}): must not end with ":".'
                )
        if not _PATH_RE.match(self.path):
            raise ValueError(f"Invalid path ({self.path}).")
        if self.tag and not is_tag(self.tag):
            raise ValueError(f"Invalid tag ({self.tag}).")
        if self.digest and not is_digest(self.digest):
            raise ValueError(f"Invalid digest ({self.digest}).")
        return self

    def combine(self) -> str:
        parts = []
        if self.registry:
            parts.append(self.registry)
            if self.path:
                parts.append("/")
        parts.append(self.path)
        if self.tag:
            parts.append(":")
            parts.append(self.tag)
        if self.digest:
            parts.append("@")
            parts.append(self.digest)
        return "".join(parts)

    def normalize(self) -> ImageName:
        registry = self.registry
        path = self.path
        if registry in ("", None, "index.docker.io", "registry.hub.docker.com"):
            registry = "docker.io"
        if registry == "docker.io" and "/" not in path and path:
            path = f"library/{path}"
        return ImageName(registry, path, self.tag, self.digest)

    def get_hostname_and_port(self) -> tuple[str, int]:
        if self.registry is None:
            raise ValueError(
                "Cannot get hostname when there is no registry. Normalize first!"
            )
        if self.registry == "docker.io":
            return "index.docker.io", 443
        parts = self.registry.split(":", 1)
        if len(parts) == 2:
            try:
                port = int(parts[1])
            except (TypeError, ValueError) as exc:
                raise ValueError(f"Cannot parse port {parts[1]!r}") from exc
            return parts[0], port
        return self.registry, 443



# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Basic documentation fragment without account data
    BASIC = r"""
notes:
  - Although the defaults are chosen so that the module can be used with the L(Let's Encrypt,https://letsencrypt.org/) CA,
    the module can in principle be used with any CA providing an ACME endpoint.
  - So far, the ACME modules have only been tested by the developers against Let's Encrypt (staging and production),
    ZeroSSL (production), and L(Pebble testing server,https://github.com/letsencrypt/Pebble).
    We have got community feedback that they also work with Sectigo ACME Service for InCommon and with HARICA.
    If you experience problems with another ACME server, please
    L(create an issue, https://github.com/ansible-collections/community.crypto/issues/new/choose)
    to help us supporting it. Feedback that an ACME server not mentioned does work is also appreciated.
requirements:
  - either C(openssl)
  - or L(cryptography,https://cryptography.io/) >= 3.3
options:
  acme_version:
    description:
      - The ACME version of the endpoint.
      - Must be V(2) for standardized ACME v2 endpoints.
      - The value V(1) is no longer supported since community.crypto 3.0.0.
    type: int
    default: 2
    choices:
      - 2
  acme_directory:
    description:
      - The ACME directory to use. This is the entry point URL to access the ACME CA server API.
      - For safety reasons the default is set to the Let's Encrypt staging server (for the ACME v1 protocol). This will create
        technically correct, but untrusted certificates.
      - "For Let's Encrypt, all staging endpoints can be found here: U(https://letsencrypt.org/docs/staging-environment/)."
      - For B(Let's Encrypt), the production directory URL for ACME v2 is U(https://acme-v02.api.letsencrypt.org/directory).
      - For B(ZeroSSL), the production directory URL for ACME v2 is U(https://acme.zerossl.com/v2/DV90).
      - For B(Sectigo), the production directory URL for ACME v2 is U(https://acme-qa.secure.trust-provider.com/v2/DV).
      - For B(HARICA), the production directory URL for ACME v2 is U(https://acme.harica.gr/XXX/directory) with XXX being specific to your account.
      - The notes for this module contain a list of ACME services this module has been tested against.
    required: true
    type: str
  validate_certs:
    description:
      - Whether calls to the ACME directory will validate TLS certificates.
      - B(Warning:) Should B(only ever) be set to V(false) for testing purposes, for example when testing against a local
        Pebble server.
    type: bool
    default: true
  select_crypto_backend:
    description:
      - Determines which crypto backend to use.
      - The default choice is V(auto), which tries to use C(cryptography) if available, and falls back to C(openssl).
      - If set to V(openssl), will try to use the C(openssl) binary.
      - If set to V(cryptography), will try to use the L(cryptography,https://cryptography.io/) library.
    type: str
    default: auto
    choices: [auto, cryptography, openssl]
  request_timeout:
    description:
      - The time Ansible should wait for a response from the ACME API.
      - This timeout is applied to all HTTP(S) requests (HEAD, GET, POST).
    type: int
    default: 10
    version_added: 2.3.0
"""

    # Account data documentation fragment
    ACCOUNT = r"""
notes:
  - If a new enough version of the C(cryptography) library is available (see Requirements for details), it will be used instead
    of the C(openssl) binary. This can be explicitly disabled or enabled with the O(select_crypto_backend) option. Note that
    using the C(openssl) binary will be slower and less secure, as private key contents always have to be stored on disk (see
    O(account_key_content)).
options:
  account_key_src:
    description:
      - Path to a file containing the ACME account RSA or Elliptic Curve key.
      - "For Elliptic Curve keys only the following curves are supported: V(secp256r1), V(secp384r1), and V(secp521r1)."
      - 'Private keys can be created with the M(community.crypto.openssl_privatekey) or M(community.crypto.openssl_privatekey_pipe)
        modules. If the requisite (cryptography) is not available, keys can also be created directly with the C(openssl) command
        line tool: RSA keys can be created with C(openssl genrsa ...). Elliptic curve keys can be created with C(openssl ecparam
        -genkey ...). Any other tool creating private keys in PEM format can be used as well.'
      - Mutually exclusive with O(account_key_content).
      - Required if O(account_key_content) is not used.
    type: path
    aliases:
      - account_key
  account_key_content:
    description:
      - Content of the ACME account RSA or Elliptic Curve key.
      - "For Elliptic Curve keys only the following curves are supported: V(secp256r1), V(secp384r1), and V(secp521r1)."
      - Mutually exclusive with O(account_key_src).
      - Required if O(account_key_src) is not used.
      - B(Warning:) the content will be written into a temporary file, which will be deleted by Ansible when the module completes.
        Since this is an important private key — it can be used to change the account key, or to revoke your certificates
        without knowing their private keys —, this might not be acceptable.
      - In case C(cryptography) is used, the content is not written into a temporary file. It can still happen that it is
        written to disk by Ansible in the process of moving the module with its argument to the node where it is executed.
    type: str
  account_key_passphrase:
    description:
      - Passphrase to use to decode the account key.
      - B(Note:) this is not supported by the C(openssl) backend, only by the C(cryptography) backend.
    type: str
    version_added: 1.6.0
  account_uri:
    description:
      - If specified, assumes that the account URI is as given. If the account key does not match this account, or an account
        with this URI does not exist, the module fails.
    type: str
"""

    # No account data documentation fragment
    NO_ACCOUNT = r"""
notes:
  - "If a new enough version of the C(cryptography) library
     is available (see Requirements for details), it will be used
     instead of the C(openssl) binary. This can be explicitly disabled
     or enabled with the O(select_crypto_backend) option. Note that using
     the C(openssl) binary will be slower."
options: {}
"""

    CERTIFICATE = r"""
options:
  csr:
    description:
      - File containing the CSR for the new certificate.
      - Can be created with M(community.crypto.openssl_csr).
      - The CSR may contain multiple Subject Alternate Names, but each one will lead to an individual challenge that must
        be fulfilled for the CSR to be signed.
      - 'B(Note): the private key used to create the CSR B(must not) be the account key. This is a bad idea from a security
        point of view, and the CA should not accept the CSR. The ACME server should return an error in this case.'
      - Precisely one of O(csr) or O(csr_content) must be specified.
    type: path
  csr_content:
    description:
      - Content of the CSR for the new certificate.
      - Can be created with M(community.crypto.openssl_csr_pipe).
      - The CSR may contain multiple Subject Alternate Names, but each one will lead to an individual challenge that must
        be fulfilled for the CSR to be signed.
      - 'B(Note): the private key used to create the CSR B(must not) be the account key. This is a bad idea from a security
        point of view, and the CA should not accept the CSR. The ACME server should return an error in this case.'
      - Precisely one of O(csr) or O(csr_content) must be specified.
    type: str
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
from ansible.plugins.action import ActionBase
try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator, )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.meraki.plugins.plugin_utils.meraki import (
    MERAKI,
    meraki_argument_spec,
    meraki_compare_equality2,
    get_dict_result,
)
from ansible_collections.cisco.meraki.plugins.plugin_utils.exceptions import (
    InconsistentParameters,
)

# Get common arguments specification
argument_spec = meraki_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    state=dict(type="str", default="present", choices=["present"]),
    name=dict(type="str"),
    tags=dict(type="list"),
    enabled=dict(type="bool"),
    poeEnabled=dict(type="bool"),
    type=dict(type="str"),
    vlan=dict(type="int"),
    voiceVlan=dict(type="int"),
    allowedVlans=dict(type="str"),
    isolationEnabled=dict(type="bool"),
    rstpEnabled=dict(type="bool"),
    stpGuard=dict(type="str"),
    stpPortFastTrunk=dict(type="bool"),
    linkNegotiation=dict(type="str"),
    portScheduleId=dict(type="str"),
    udld=dict(type="str"),
    accessPolicyType=dict(type="str"),
    accessPolicyNumber=dict(type="int"),
    macAllowList=dict(type="list"),
    macWhitelistLimit=dict(type="int"),
    stickyMacAllowList=dict(type="list"),
    stickyMacAllowListLimit=dict(type="int"),
    stormControlEnabled=dict(type="bool"),
    flexibleStackingEnabled=dict(type="bool"),
    daiTrusted=dict(type="bool"),
    profile=dict(type="dict"),
    dot3az=dict(type="dict"),
    highSpeed=dict(type="dict"),
    organizationId=dict(type="str"),
    configTemplateId=dict(type="str"),
    profileId=dict(type="str"),
    portId=dict(type="str"),
))

required_if = [("state", "present", ["configTemplateId", "name",
                                     "organizationId", "portId", "profileId"], True), ]
required_one_of = []
mutually_exclusive = []
required_together = []


class OrganizationsConfigTemplatesSwitchProfilesPorts(object):
    def __init__(self, params, meraki):
        self.meraki = meraki
        self.new_object = dict(
            name=params.get("name"),
            tags=params.get("tags"),
            enabled=params.get("enabled"),
            poeEnabled=params.get("poeEnabled"),
            type=params.get("type"),
            vlan=params.get("vlan"),
            voiceVlan=params.get("voiceVlan"),
            allowedVlans=params.get("allowedVlans"),
            isolationEnabled=params.get("isolationEnabled"),
            rstpEnabled=params.get("rstpEnabled"),
            stpGuard=params.get("stpGuard"),
            stpPortFastTrunk=params.get("stpPortFastTrunk"),
            linkNegotiation=params.get("linkNegotiation"),
            portScheduleId=params.get("portScheduleId"),
            udld=params.get("udld"),
            accessPolicyType=params.get("accessPolicyType"),
            accessPolicyNumber=params.get("accessPolicyNumber"),
            macAllowList=params.get("macAllowList"),
            macWhitelistLimit=params.get("macWhitelistLimit"),
            stickyMacAllowList=params.get("stickyMacAllowList"),
            stickyMacAllowListLimit=params.get("stickyMacAllowListLimit"),
            stormControlEnabled=params.get("stormControlEnabled"),
            flexibleStackingEnabled=params.get("flexibleStackingEnabled"),
            daiTrusted=params.get("daiTrusted"),
            profile=params.get("profile"),
            dot3az=params.get("dot3az"),
            highSpeed=params.get("highSpeed"),
            organizationId=params.get("organizationId"),
            config_template_id=params.get("configTemplateId"),
            profile_id=params.get("profileId"),
            port_id=params.get("portId"),
        )

    def get_all_params(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('organizationId') is not None or self.new_object.get(
                'organization_id') is not None:
            new_object_params['organizationId'] = self.new_object.get(
                'organizationId') or self.new_object.get('organization_id')
        if self.new_object.get('configTemplateId') is not None or self.new_object.get(
                'config_template_id') is not None:
            new_object_params['configTemplateId'] = self.new_object.get(
                'configTemplateId') or self.new_object.get('config_template_id')
        if self.new_object.get('profileId') is not None or self.new_object.get(
                'profile_id') is not None:
            new_object_params['profileId'] = self.new_object.get(
                'profileId') or self.new_object.get('profile_id')
        return new_object_params

    def get_params_by_id(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('organizationId') is not None or self.new_object.get(
                'organization_id') is not None:
            new_object_params['organizationId'] = self.new_object.get(
                'organizationId') or self.new_object.get('organization_id')
        if self.new_object.get('configTemplateId') is not None or self.new_object.get(
                'config_template_id') is not None:
            new_object_params['configTemplateId'] = self.new_object.get(
                'configTemplateId') or self.new_object.get('config_template_id')
        if self.new_object.get('profileId') is not None or self.new_object.get(
                'profile_id') is not None:
            new_object_params['profileId'] = self.new_object.get(
                'profileId') or self.new_object.get('profile_id')
        if self.new_object.get('portId') is not None or self.new_object.get(
                'port_id') is not None:
            new_object_params['portId'] = self.new_object.get('portId') or \
                self.new_object.get('port_id')
        return new_object_params

    def update_by_id_params(self):
        new_object_params = {}
        if self.new_object.get('name') is not None or self.new_object.get(
                'name') is not None:
            new_object_params['name'] = self.new_object.get('name') or \
                self.new_object.get('name')
        if self.new_object.get('tags') is not None or self.new_object.get(
                'tags') is not None:
            new_object_params['tags'] = self.new_object.get('tags') or \
                self.new_object.get('tags')
        if self.new_object.get('enabled') is not None or self.new_object.get(
                'enabled') is not None:
            new_object_params['enabled'] = self.new_object.get('enabled')
        if self.new_object.get('poeEnabled') is not None or self.new_object.get(
                'poe_enabled') is not None:
            new_object_params['poeEnabled'] = self.new_object.get('poeEnabled')
        if self.new_object.get('type') is not None or self.new_object.get(
                'type') is not None:
            new_object_params['type'] = self.new_object.get('type') or \
                self.new_object.get('type')
        if self.new_object.get('vlan') is not None or self.new_object.get(
                'vlan') is not None:
            new_object_params['vlan'] = self.new_object.get('vlan') or \
                self.new_object.get('vlan')
        if self.new_object.get('voiceVlan') is not None or self.new_object.get(
                'voice_vlan') is not None:
            new_object_params['voiceVlan'] = self.new_object.get(
                'voiceVlan') or self.new_object.get('voice_vlan')
        if self.new_object.get('allowedVlans') is not None or self.new_object.get(
                'allowed_vlans') is not None:
            new_object_params['allowedVlans'] = self.new_object.get(
                'allowedVlans') or self.new_object.get('allowed_vlans')
        if self.new_object.get('isolationEnabled') is not None or self.new_object.get(
                'isolation_enabled') is not None:
            new_object_params['isolationEnabled'] = self.new_object.get(
                'isolationEnabled')
        if self.new_object.get('rstpEnabled') is not None or self.new_object.get(
                'rstp_enabled') is not None:
            new_object_params['rstpEnabled'] = self.new_object.get(
                'rstpEnabled')
        if self.new_object.get('stpGuard') is not None or self.new_object.get(
                'stp_guard') is not None:
            new_object_params['stpGuard'] = self.new_object.get(
                'stpGuard') or self.new_object.get('stp_guard')
        if self.new_object.get('stpPortFastTrunk') is not None or self.new_object.get(
                'stp_port_fast_trunk') is not None:
            new_object_params['stpPortFastTrunk'] = self.new_object.get(
                'stpPortFastTrunk')
        if self.new_object.get('linkNegotiation') is not None or self.new_object.get(
                'link_negotiation') is not None:
            new_object_params['linkNegotiation'] = self.new_object.get(
                'linkNegotiation') or self.new_object.get('link_negotiation')
        if self.new_object.get('portScheduleId') is not None or self.new_object.get(
                'port_schedule_id') is not None:
            new_object_params['portScheduleId'] = self.new_object.get(
                'portScheduleId') or self.new_object.get('port_schedule_id')
        if self.new_object.get('udld') is not None or self.new_object.get(
                'udld') is not None:
            new_object_params['udld'] = self.new_object.get('udld') or \
                self.new_object.get('udld')
        if self.new_object.get('accessPolicyType') is not None or self.new_object.get(
                'access_policy_type') is not None:
            new_object_params['accessPolicyType'] = self.new_object.get(
                'accessPolicyType') or self.new_object.get('access_policy_type')
        if self.new_object.get('accessPolicyNumber') is not None or self.new_object.get(
                'access_policy_number') is not None:
            new_object_params['accessPolicyNumber'] = self.new_object.get(
                'accessPolicyNumber') or self.new_object.get('access_policy_number')
        if self.new_object.get('macAllowList') is not None or self.new_object.get(
                'mac_allow_list') is not None:
            new_object_params['macAllowList'] = self.new_object.get(
                'macAllowList') or self.new_object.get('mac_allow_list')
        if self.new_object.get('macWhitelistLimit') is not None or self.new_object.get(
                'mac_whitelist_limit') is not None:
            new_object_params['macWhitelistLimit'] = self.new_object.get(
                'macWhitelistLimit') or self.new_object.get('mac_whitelist_limit')
        if self.new_object.get('stickyMacAllowList') is not None or self.new_object.get(
                'sticky_mac_allow_list') is not None:
            new_object_params['stickyMacAllowList'] = self.new_object.get(
                'stickyMacAllowList') or self.new_object.get('sticky_mac_allow_list')
        if self.new_object.get('stickyMacAllowListLimit') is not None or self.new_object.get(
                'sticky_mac_allow_list_limit') is not None:
            new_object_params['stickyMacAllowListLimit'] = self.new_object.get(
                'stickyMacAllowListLimit') or self.new_object.get('sticky_mac_allow_list_limit')
        if self.new_object.get('stormControlEnabled') is not None or self.new_object.get(
                'storm_control_enabled') is not None:
            new_object_params['stormControlEnabled'] = self.new_object.get(
                'stormControlEnabled')
        if self.new_object.get('flexibleStackingEnabled') is not None or self.new_object.get(
                'flexible_stacking_enabled') is not None:
            new_object_params['flexibleStackingEnabled'] = self.new_object.get(
                'flexibleStackingEnabled')
        if self.new_object.get('daiTrusted') is not None or self.new_object.get(
                'dai_trusted') is not None:
            new_object_params['daiTrusted'] = self.new_object.get('daiTrusted')
        if self.new_object.get('profile') is not None or self.new_object.get(
                'profile') is not None:
            new_object_params['profile'] = self.new_object.get('profile') or \
                self.new_object.get('profile')
        if self.new_object.get('dot3az') is not None or self.new_object.get(
                'dot3az') is not None:
            new_object_params['dot3az'] = self.new_object.get('dot3az') or \
                self.new_object.get('dot3az')
        if self.new_object.get('highSpeed') is not None or self.new_object.get(
                'high_speed') is not None:
            new_object_params['highSpeed'] = self.new_object.get(
                'highSpeed') or self.new_object.get('high_speed')
        if self.new_object.get('organizationId') is not None or self.new_object.get(
                'organization_id') is not None:
            new_object_params['organizationId'] = self.new_object.get(
                'organizationId') or self.new_object.get('organization_id')
        if self.new_object.get('configTemplateId') is not None or self.new_object.get(
                'config_template_id') is not None:
            new_object_params['configTemplateId'] = self.new_object.get(
                'configTemplateId') or self.new_object.get('config_template_id')
        if self.new_object.get('profileId') is not None or self.new_object.get(
                'profile_id') is not None:
            new_object_params['profileId'] = self.new_object.get(
                'profileId') or self.new_object.get('profile_id')
        if self.new_object.get('portId') is not None or self.new_object.get(
                'port_id') is not None:
            new_object_params['portId'] = self.new_object.get('portId') or \
                self.new_object.get('port_id')
        return new_object_params

    def get_object_by_name(self, name):
        result = None
        # NOTE: Does not have a get by name method, using get all
        try:
            items = self.meraki.exec_meraki(
                family="switch",
                function="getOrganizationConfigTemplateSwitchProfilePorts",
                params=self.get_all_params(name=name),
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'name', name)
            if result is None:
                result = items
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def get_object_by_id(self, id):
        result = None
        try:
            items = self.meraki.exec_meraki(
                family="switch",
                function="getOrganizationConfigTemplateSwitchProfilePort",
                params=self.get_params_by_id()
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = items
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def exists(self):
        prev_obj = None
        id_exists = False
        name_exists = False
        o_id = self.new_object.get("id")
        o_id = o_id or self.new_object.get(
            "port_id") or self.new_object.get("portId")
        name = self.new_object.get("name")
        if o_id:
            prev_obj = self.get_object_by_id(o_id)
            id_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if not id_exists and name:
            prev_obj = self.get_object_by_name(name)
            name_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if name_exists:
            _id = prev_obj.get("id")
            _id = _id or prev_obj.get("portId")
            if id_exists and name_exists and o_id != _id:
                raise InconsistentParameters(
                    "The 'id' and 'name' params don't refer to the same object")
            if _id:
                self.new_object.update(dict(id=_id))
                self.new_object.update(dict(portId=_id))
            if _id:
                prev_obj = self.get_object_by_id(_id)
        it_exists = prev_obj is not None and isinstance(prev_obj, dict)
        return (it_exists, prev_obj)

    def requires_update(self, current_obj):
        requested_obj = self.new_object

        obj_params = [
            ("name", "name"),
            ("tags", "tags"),
            ("enabled", "enabled"),
            ("poeEnabled", "poeEnabled"),
            ("type", "type"),
            ("vlan", "vlan"),
            ("voiceVlan", "voiceVlan"),
            ("allowedVlans", "allowedVlans"),
            ("isolationEnabled", "isolationEnabled"),
            ("rstpEnabled", "rstpEnabled"),
            ("stpGuard", "stpGuard"),
            ("stpPortFastTrunk", "stpPortFastTrunk"),
            ("linkNegotiation", "linkNegotiation"),
            ("portScheduleId", "portScheduleId"),
            ("udld", "udld"),
            ("accessPolicyType", "accessPolicyType"),
            ("accessPolicyNumber", "accessPolicyNumber"),
            ("macAllowList", "macAllowList"),
            ("macWhitelistLimit", "macWhitelistLimit"),
            ("stickyMacAllowList", "stickyMacAllowList"),
            ("stickyMacAllowListLimit", "stickyMacAllowListLimit"),
            ("stormControlEnabled", "stormControlEnabled"),
            ("flexibleStackingEnabled", "flexibleStackingEnabled"),
            ("daiTrusted", "daiTrusted"),
            ("profile", "profile"),
            ("dot3az", "dot3az"),
            ("highSpeed", "highSpeed"), ("profileId", "profileId"),
            ("portId", "portId"),
        ]
        # Method 1. Params present in request (Ansible) obj are the same as the current (ISE) params
        # If any does not have eq params, it requires update
        return any(
            not meraki_compare_equality2(
                current_obj.get(meraki_param),
                requested_obj.get(ansible_param)) for (
                meraki_param,
                ansible_param) in obj_params)

    def update(self):
        id = self.new_object.get("id")
        id = id or self.new_object.get("portId")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
                id_ = id_ or prev_obj_name.get("portId")
            if id_:
                self.new_object.update(dict(portid=id_))
        result = self.meraki.exec_meraki(
            family="switch",
            function="updateOrganizationConfigTemplateSwitchProfilePort",
            params=self.update_by_id_params(),
            op_modifies=True,
        )
        return result


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail(
                "ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = False
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        meraki = MERAKI(self._task.args)
        obj = OrganizationsConfigTemplatesSwitchProfilesPorts(
            self._task.args, meraki)

        state = self._task.args.get("state")

        response = None
        if state == "present":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                if obj.requires_update(prev_obj):
                    response = obj.update()
                    meraki.object_updated()
                else:
                    response = prev_obj
                    meraki.object_already_present()
            else:
                meraki.fail_json(
                    "Object does not exists, plugin only has update")

        self._result.update(dict(meraki_response=response))
        self._result.update(meraki.exit_json())
        return self._result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Tim Cragg (@timcragg) <tcragg@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_management_network_instance_profile
version_added: "2.13.0"
short_description: Manage external management network instance profiles (mgmt:InstP).
description:
- Manage external management network instance profiles on Cisco ACI fabrics.
options:
  profile:
    description:
    - The name of the external management network instance profile.
    type: str
    aliases: [ name, profile_name ]
  qos_class:
    description:
    - QoS priority class identifier.
    - The APIC defaults to C(unspecified) when unset during creation.
    type: str
    aliases: [ qos, priority, prio ]
    choices: [ level1, level2, level3, level4, level5, level6, unspecified ]
  description:
    description:
    - The description for the external management network instance profile.
    type: str
    aliases: [ descr ]
  subnets:
    description:
    - The list of subnets in CIDR format to associate with the external management network instance profile.
    - When state is C(present) and a list of subnets is provided, any existing subnets will be removed from the object if they are not present in this list.
    - When subnets is set to an empty list and state is C(present), all existing subnets will be removed from the profile.
    type: list
    elements: str
    aliases: [ subnet_list, networks, network_list ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(mgmt:InstP).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Tim Cragg (@timcragg)
"""

EXAMPLES = r"""
- name: Add a new external management network instance profile
  cisco.aci.aci_management_network_instance_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: lab_network_inst_profile
    subnets:
      - 10.20.30.0/24
      - 192.168.10.0/24
    state: present
  delegate_to: localhost

- name: Delete all subnets from management network instance profile
  cisco.aci.aci_management_network_instance_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: lab_network_inst_profile
    subnets: []
    state: present
  delegate_to: localhost

- name: Remove an external management network instance profile
  cisco.aci.aci_management_network_instance_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: lab_network_inst_profile
    state: absent
  delegate_to: localhost

- name: Query an external management network instance profile
  cisco.aci.aci_management_network_instance_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: lab_network_inst_profile
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all external management network instance profiles
  cisco.aci.aci_management_network_instance_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import VALID_QOS_CLASSES


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        profile=dict(type="str", aliases=["name", "profile_name"]),
        qos_class=dict(type="str", aliases=["qos", "priority", "prio"], choices=VALID_QOS_CLASSES),
        description=dict(type="str", aliases=["descr"]),
        subnets=dict(type="list", elements="str", aliases=["subnet_list", "networks", "network_list"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        name_alias=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["profile"]],
            ["state", "present", ["profile"]],
        ],
    )

    profile = module.params.get("profile")
    qos_class = module.params.get("qos_class")
    description = module.params.get("description")
    subnets = module.params.get("subnets")
    state = module.params.get("state")
    name_alias = module.params.get("name_alias")

    aci = ACIModule(module)
    aci.construct_url(
        root_class=dict(
            aci_class="mgmtInstP",
            aci_rn="tn-mgmt/extmgmt-default/instp-{0}".format(profile),
            module_object=profile,
            target_filter={"name": profile},
        ),
        child_classes=["mgmtSubnet"],
    )

    aci.get_existing()

    if state == "present":
        child_configs = []
        if subnets is not None:
            for subnet in subnets:
                child_configs.append(dict(mgmtSubnet=dict(attributes=dict(ip=subnet))))
            if isinstance(aci.existing, list) and len(aci.existing) > 0:
                for child in aci.existing[0].get("mgmtInstP", {}).get("children", []):
                    # Remove any existing subnet entries that are not in the requested subnet list
                    if child.get("mgmtSubnet") and child.get("mgmtSubnet").get("attributes", {}).get("ip") not in subnets:
                        child_configs.append(
                            {
                                "mgmtSubnet": {
                                    "attributes": {
                                        "ip": child.get("mgmtSubnet").get("attributes", {}).get("ip"),
                                        "status": "deleted",
                                    }
                                }
                            }
                        )

        aci.payload(
            aci_class="mgmtInstP",
            class_config=dict(
                descr=description,
                name=profile,
                nameAlias=name_alias,
                prio=qos_class,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class="mgmtInstP")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (C) 2019 Huawei
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

###############################################################################
# Documentation
###############################################################################

DOCUMENTATION = r"""
module: hwc_vpc_subnet
description:
  - Subnet management.
short_description: Creates a resource of VPC/Subnet in Huawei Cloud
version_added: '0.2.0'
author: Huawei Inc. (@huaweicloud)
requirements:
  - keystoneauth1 >= 3.6.0
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether the given object should exist in Huawei Cloud.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  timeouts:
    description:
      - The timeouts for each operations.
    type: dict
    default: {}
    suboptions:
      create:
        description:
          - The timeouts for create operation.
        type: str
        default: '15m'
      update:
        description:
          - The timeouts for update operation.
        type: str
        default: '15m'
  cidr:
    description:
      - Specifies the subnet CIDR block. The value must be within the VPC CIDR block and be in CIDR format. The subnet mask
        cannot be greater than 28. Cannot be changed after creating the subnet.
    type: str
    required: true
  gateway_ip:
    description:
      - Specifies the gateway of the subnet. The value must be an IP address in the subnet. Cannot be changed after creating
        the subnet.
    type: str
    required: true
  name:
    description:
      - Specifies the subnet name. The value is a string of 1 to 64 characters that can contain letters, digits, underscores
        (V(_)), hyphens (V(-)), and periods (V(.)).
    type: str
    required: true
  vpc_id:
    description:
      - Specifies the ID of the VPC to which the subnet belongs. Cannot be changed after creating the subnet.
    type: str
    required: true
  availability_zone:
    description:
      - Specifies the AZ to which the subnet belongs. Cannot be changed after creating the subnet.
    type: str
  dhcp_enable:
    description:
      - Specifies whether DHCP is enabled for the subnet. The value can be true (enabled) or false(disabled), and default
        value is true. If this parameter is set to false, newly created ECSs cannot obtain IP addresses, and usernames and
        passwords cannot be injected using Cloud-init.
    type: bool
  dns_address:
    description:
      - Specifies the DNS server addresses for subnet. The address in the head is used first.
    type: list
    elements: str
extends_documentation_fragment:
  - community.general._hwc
  - community.general._attributes
"""

EXAMPLES = r"""
# create subnet
- name: Create vpc
  hwc_network_vpc:
    cidr: "192.168.100.0/24"
    name: "ansible_network_vpc_test"
  register: vpc
- name: Create subnet
  community.general.hwc_vpc_subnet:
    vpc_id: "{{ vpc.id }}"
    cidr: "192.168.100.0/26"
    gateway_ip: "192.168.100.32"
    name: "ansible_network_subnet_test"
    dhcp_enable: true
"""

RETURN = r"""
cidr:
  description:
    - Specifies the subnet CIDR block. The value must be within the VPC CIDR block and be in CIDR format. The subnet mask
      cannot be greater than 28.
  type: str
  returned: success
gateway_ip:
  description:
    - Specifies the gateway of the subnet. The value must be an IP address in the subnet.
  type: str
  returned: success
name:
  description:
    - Specifies the subnet name. The value is a string of 1 to 64 characters that can contain letters, digits, underscores
      (V(_)), hyphens (V(-)), and periods (V(.)).
  type: str
  returned: success
vpc_id:
  description:
    - Specifies the ID of the VPC to which the subnet belongs.
  type: str
  returned: success
availability_zone:
  description:
    - Specifies the AZ to which the subnet belongs.
  type: str
  returned: success
dhcp_enable:
  description:
    - Specifies whether DHCP is enabled for the subnet. The value can be true (enabled) or false(disabled), and default value
      is true. If this parameter is set to false, newly created ECSs cannot obtain IP addresses, and usernames and passwords
      cannot be injected using Cloud-init.
  type: bool
  returned: success
dns_address:
  description:
    - Specifies the DNS server addresses for subnet. The address in the head is used first.
  type: list
  returned: success
"""

from ansible_collections.community.general.plugins.module_utils._hwc_utils import (
    Config,
    HwcClientException,
    HwcClientException404,
    HwcModule,
    are_different_dicts,
    build_path,
    get_region,
    is_empty_value,
    navigate_value,
    wait_to_finish,
)


def build_module():
    return HwcModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"], type="str"),
            timeouts=dict(
                type="dict",
                options=dict(
                    create=dict(default="15m", type="str"),
                    update=dict(default="15m", type="str"),
                ),
                default=dict(),
            ),
            cidr=dict(type="str", required=True),
            gateway_ip=dict(type="str", required=True),
            name=dict(type="str", required=True),
            vpc_id=dict(type="str", required=True),
            availability_zone=dict(type="str"),
            dhcp_enable=dict(type="bool"),
            dns_address=dict(type="list", elements="str"),
        ),
        supports_check_mode=True,
    )


def main():
    """Main function"""

    module = build_module()
    config = Config(module, "vpc")

    try:
        resource = None
        if module.params.get("id"):
            resource = True
        else:
            v = search_resource(config)
            if len(v) > 1:
                raise Exception(f"Found more than one resource({', '.join([navigate_value(i, ['id']) for i in v])})")

            if len(v) == 1:
                resource = v[0]
                module.params["id"] = navigate_value(resource, ["id"])

        result = {}
        changed = False
        if module.params["state"] == "present":
            if resource is None:
                if not module.check_mode:
                    create(config)
                changed = True

            current = read_resource(config, exclude_output=True)
            expect = user_input_parameters(module)
            if are_different_dicts(expect, current):
                if not module.check_mode:
                    update(config)
                changed = True

            result = read_resource(config)
            result["id"] = module.params.get("id")
        else:
            if resource:
                if not module.check_mode:
                    delete(config)
                changed = True

    except Exception as ex:
        module.fail_json(msg=str(ex))

    else:
        result["changed"] = changed
        module.exit_json(**result)


def user_input_parameters(module):
    return {
        "availability_zone": module.params.get("availability_zone"),
        "cidr": module.params.get("cidr"),
        "dhcp_enable": module.params.get("dhcp_enable"),
        "dns_address": module.params.get("dns_address"),
        "gateway_ip": module.params.get("gateway_ip"),
        "name": module.params.get("name"),
        "vpc_id": module.params.get("vpc_id"),
    }


def create(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    opts = user_input_parameters(module)

    params = build_create_parameters(opts)
    r = send_create_request(module, params, client)
    obj = async_wait_create(config, r, client, timeout)
    module.params["id"] = navigate_value(obj, ["subnet", "id"])


def update(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    timeout = 60 * int(module.params["timeouts"]["update"].rstrip("m"))
    opts = user_input_parameters(module)

    params = build_update_parameters(opts)
    if params:
        r = send_update_request(module, params, client)
        async_wait_update(config, r, client, timeout)


def delete(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    send_delete_request(module, None, client)

    url = build_path(module, "subnets/{id}")

    def _refresh_status():
        try:
            client.get(url)
        except HwcClientException404:
            return True, "Done"

        except Exception:
            return None, ""

        return True, "Pending"

    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    try:
        wait_to_finish(["Done"], ["Pending"], _refresh_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_vpc_subnet): error waiting for api(delete) to be done, error= {ex}")


def read_resource(config, exclude_output=False):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    res = {}

    r = send_read_request(module, client)
    res["read"] = fill_read_resp_body(r)

    return update_properties(module, res, None, exclude_output)


def _build_query_link(opts):
    query_link = "?marker={marker}&limit=10"
    v = navigate_value(opts, ["vpc_id"])
    if v:
        query_link += f"&vpc_id={v}"

    return query_link


def search_resource(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    opts = user_input_parameters(module)
    identity_obj = _build_identity_object(opts)
    query_link = _build_query_link(opts)
    link = f"subnets{query_link}"

    result = []
    p = {"marker": ""}
    while True:
        url = link.format(**p)
        r = send_list_request(module, client, url)
        if not r:
            break

        for item in r:
            item = fill_list_resp_body(item)
            if not are_different_dicts(identity_obj, item):
                result.append(item)

        if len(result) > 1:
            break

        p["marker"] = r[-1].get("id")

    return result


def build_create_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["availability_zone"], None)
    if not is_empty_value(v):
        params["availability_zone"] = v

    v = navigate_value(opts, ["cidr"], None)
    if not is_empty_value(v):
        params["cidr"] = v

    v = navigate_value(opts, ["dhcp_enable"], None)
    if v is not None:
        params["dhcp_enable"] = v

    v = expand_create_dns_list(opts, None)
    if not is_empty_value(v):
        params["dnsList"] = v

    v = navigate_value(opts, ["gateway_ip"], None)
    if not is_empty_value(v):
        params["gateway_ip"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = expand_create_primary_dns(opts, None)
    if not is_empty_value(v):
        params["primary_dns"] = v

    v = expand_create_secondary_dns(opts, None)
    if not is_empty_value(v):
        params["secondary_dns"] = v

    v = navigate_value(opts, ["vpc_id"], None)
    if not is_empty_value(v):
        params["vpc_id"] = v

    if not params:
        return params

    params = {"subnet": params}

    return params


def expand_create_dns_list(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    return v if (v and len(v) > 2) else []


def expand_create_primary_dns(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    return v[0] if v else ""


def expand_create_secondary_dns(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    return v[1] if (v and len(v) > 1) else ""


def send_create_request(module, params, client):
    url = "subnets"
    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_subnet): error running api(create), error: {ex}"
        module.fail_json(msg=msg)

    return r


def async_wait_create(config, result, client, timeout):
    module = config.module

    path_parameters = {
        "subnet_id": ["subnet", "id"],
    }
    data = {key: navigate_value(result, path) for key, path in path_parameters.items()}

    url = build_path(module, "subnets/{subnet_id}", data)

    def _query_status():
        r = None
        try:
            r = client.get(url, timeout=timeout)
        except HwcClientException:
            return None, ""

        try:
            s = navigate_value(r, ["subnet", "status"])
            return r, s
        except Exception:
            return None, ""

    try:
        return wait_to_finish(["ACTIVE"], ["UNKNOWN"], _query_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_vpc_subnet): error waiting for api(create) to be done, error= {ex}")


def build_update_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["dhcp_enable"], None)
    if v is not None:
        params["dhcp_enable"] = v

    v = expand_update_dns_list(opts, None)
    if v is not None:
        params["dnsList"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = expand_update_primary_dns(opts, None)
    if v is not None:
        params["primary_dns"] = v

    v = expand_update_secondary_dns(opts, None)
    if v is not None:
        params["secondary_dns"] = v

    if not params:
        return params

    params = {"subnet": params}

    return params


def expand_update_dns_list(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    if v:
        if len(v) > 2:
            return v
        return None
    return []


def expand_update_primary_dns(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    return v[0] if v else ""


def expand_update_secondary_dns(d, array_index):
    v = navigate_value(d, ["dns_address"], array_index)
    return v[1] if (v and len(v) > 1) else ""


def send_update_request(module, params, client):
    url = build_path(module, "vpcs/{vpc_id}/subnets/{id}")

    try:
        r = client.put(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_subnet): error running api(update), error: {ex}"
        module.fail_json(msg=msg)

    return r


def async_wait_update(config, result, client, timeout):
    module = config.module

    path_parameters = {
        "subnet_id": ["subnet", "id"],
    }
    data = {key: navigate_value(result, path) for key, path in path_parameters.items()}

    url = build_path(module, "subnets/{subnet_id}", data)

    def _query_status():
        r = None
        try:
            r = client.get(url, timeout=timeout)
        except HwcClientException:
            return None, ""

        try:
            s = navigate_value(r, ["subnet", "status"])
            return r, s
        except Exception:
            return None, ""

    try:
        return wait_to_finish(["ACTIVE"], ["UNKNOWN"], _query_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_vpc_subnet): error waiting for api(update) to be done, error= {ex}")


def send_delete_request(module, params, client):
    url = build_path(module, "vpcs/{vpc_id}/subnets/{id}")

    try:
        r = client.delete(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_subnet): error running api(delete), error: {ex}"
        module.fail_json(msg=msg)

    return r


def send_read_request(module, client):
    url = build_path(module, "subnets/{id}")

    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_subnet): error running api(read), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["subnet"], None)


def fill_read_resp_body(body):
    result = dict()

    result["availability_zone"] = body.get("availability_zone")

    result["cidr"] = body.get("cidr")

    result["dhcp_enable"] = body.get("dhcp_enable")

    result["dnsList"] = body.get("dnsList")

    result["gateway_ip"] = body.get("gateway_ip")

    result["id"] = body.get("id")

    result["name"] = body.get("name")

    result["neutron_network_id"] = body.get("neutron_network_id")

    result["neutron_subnet_id"] = body.get("neutron_subnet_id")

    result["primary_dns"] = body.get("primary_dns")

    result["secondary_dns"] = body.get("secondary_dns")

    result["status"] = body.get("status")

    result["vpc_id"] = body.get("vpc_id")

    return result


def update_properties(module, response, array_index, exclude_output=False):
    r = user_input_parameters(module)

    v = navigate_value(response, ["read", "availability_zone"], array_index)
    r["availability_zone"] = v

    v = navigate_value(response, ["read", "cidr"], array_index)
    r["cidr"] = v

    v = navigate_value(response, ["read", "dhcp_enable"], array_index)
    r["dhcp_enable"] = v

    v = navigate_value(response, ["read", "dnsList"], array_index)
    r["dns_address"] = v

    v = navigate_value(response, ["read", "gateway_ip"], array_index)
    r["gateway_ip"] = v

    v = navigate_value(response, ["read", "name"], array_index)
    r["name"] = v

    v = navigate_value(response, ["read", "vpc_id"], array_index)
    r["vpc_id"] = v

    return r


def send_list_request(module, client, url):
    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_subnet): error running api(list), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["subnets"], None)


def _build_identity_object(all_opts):
    result = dict()

    v = navigate_value(all_opts, ["availability_zone"], None)
    result["availability_zone"] = v

    v = navigate_value(all_opts, ["cidr"], None)
    result["cidr"] = v

    v = navigate_value(all_opts, ["dhcp_enable"], None)
    result["dhcp_enable"] = v

    v = navigate_value(all_opts, ["dns_address"], None)
    result["dnsList"] = v

    v = navigate_value(all_opts, ["gateway_ip"], None)
    result["gateway_ip"] = v

    result["id"] = None

    v = navigate_value(all_opts, ["name"], None)
    result["name"] = v

    result["neutron_network_id"] = None

    result["neutron_subnet_id"] = None

    result["primary_dns"] = None

    result["secondary_dns"] = None

    result["status"] = None

    v = navigate_value(all_opts, ["vpc_id"], None)
    result["vpc_id"] = v

    return result


def fill_list_resp_body(body):
    result = dict()

    result["availability_zone"] = body.get("availability_zone")

    result["cidr"] = body.get("cidr")

    result["dhcp_enable"] = body.get("dhcp_enable")

    result["dnsList"] = body.get("dnsList")

    result["gateway_ip"] = body.get("gateway_ip")

    result["id"] = body.get("id")

    result["name"] = body.get("name")

    result["neutron_network_id"] = body.get("neutron_network_id")

    result["neutron_subnet_id"] = body.get("neutron_subnet_id")

    result["primary_dns"] = body.get("primary_dns")

    result["secondary_dns"] = body.get("secondary_dns")

    result["status"] = body.get("status")

    result["vpc_id"] = body.get("vpc_id")

    return result


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: luks_device

short_description: Manage encrypted (LUKS) devices

description:
  - Module manages L(LUKS,https://en.wikipedia.org/wiki/Linux_Unified_Key_Setup) on given device. Supports creating, destroying,
    opening and closing of LUKS container and adding or removing new keys and passphrases.
extends_documentation_fragment:
  - community.crypto._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  idempotent:
    support: full

options:
  device:
    description:
      - Device to work with (for example V(/dev/sda1)). Needed in most cases. Can be omitted only when O(state=closed) together
        with O(name) is provided.
    type: str
  state:
    description:
      - Desired state of the LUKS container. Based on its value creates, destroys, opens or closes the LUKS container on a
        given device.
      - V(present) will create LUKS container unless already present. Requires O(device) and either O(keyfile) or O(passphrase)
        options to be provided.
      - V(absent) will remove existing LUKS container if it exists. Requires O(device) or O(name) to be specified.
      - V(opened) will unlock the LUKS container. Requires O(device) and one of O(keyfile), O(passphrase), O(tpm2_device) to be specified.
        If the container does not exist it will be created first, however O(tpm2_device) can not be used for creation.
        Use the O(name) option to set the name of the opened container. Otherwise the name will be generated automatically and returned as a part of the result.
      - V(closed) will lock the LUKS container. However if the container does not exist it will be created. Requires O(device)
        and either O(keyfile) or O(passphrase) options to be provided. If container does already exist O(device) or O(name)
        will suffice.
    type: str
    default: present
    choices: [present, absent, opened, closed]
  name:
    description:
      - Sets container name when O(state=opened). Can be used instead of O(device) when closing the existing container (that
        is, when O(state=closed)).
    type: str
  keyfile:
    description:
      - Used to unlock the container. Either a O(keyfile) or a O(passphrase) is needed for most of the operations. Parameter
        value is the path to the keyfile with the passphrase.
      - BEWARE that working with keyfiles in plaintext is dangerous. Make sure that they are protected.
    type: path
  passphrase:
    description:
      - Used to unlock the container. Either a O(passphrase) or a O(keyfile) is needed for most of the operations. Parameter
        value is a string with the passphrase.
      - B(Note) that the passphrase must be UTF-8 encoded text. If you want to use arbitrary binary data, or text using
        another encoding, use the O(passphrase_encoding) option and provide the passphrase Base64 encoded.
    type: str
    version_added: '1.0.0'
  tpm2_device:
    description:
      - Used to unlock the container, but can not be used for container creation. A device node path referring to a TPM2 chip (for example V(/dev/tpmrm0)).
        Alternatively the special value V(auto) may be specified, in order to automatically determine the device node of a currently
        discovered TPM2 device (of which there must be exactly one).
      - B(Note) that only LUKS2 containers are supported
      - B(Note) that systemd-cryptsetup (v256 or newer) is required.
    type: str
    version_added: '3.1.0'
  passphrase_encoding:
    description:
      - Determine how passphrases are provided to parameters such as O(passphrase), O(new_passphrase), and O(remove_passphrase).
    type: str
    default: text
    choices:
      text:
        - The passphrase is provided as UTF-8 encoded text.
      base64:
        - The passphrase is provided as Base64 encoded bytes.
        - Use the P(ansible.builtin.b64encode#filter) filter to Base64-encode binary data.
    version_added: 2.23.0
  keyslot:
    description:
      - Adds the O(keyfile) or O(passphrase) to a specific keyslot when creating a new container on O(device). Parameter value
        is the number of the keyslot.
      - Defines the keyslot whose priority will be changed by O(keyslot_priority)
      - B(Note) that a device of O(type=luks1) supports the keyslot numbers V(0)-V(7) and a device of O(type=luks2) supports
        the keyslot numbers V(0)-V(31). In order to use the keyslots V(8)-V(31) when creating a new container, setting O(type)
        to V(luks2) is required.
    type: int
    version_added: '2.16.0'
  keysize:
    description:
      - Sets the key size only if LUKS container does not exist.
    type: int
    version_added: '1.0.0'
  new_keyfile:
    description:
      - Adds additional key to given container on O(device). Needs O(keyfile) or O(passphrase) option for authorization. LUKS
        container supports up to 8 keyslots. Parameter value is the path to the keyfile with the passphrase.
      - NOTE that adding additional keys is idempotent only since community.crypto 1.4.0. For older versions, a new keyslot
        will be used even if another keyslot already exists for this keyfile.
      - BEWARE that working with keyfiles in plaintext is dangerous. Make sure that they are protected.
    type: path
  new_passphrase:
    description:
      - Adds additional passphrase to given container on O(device). Needs O(keyfile) or O(passphrase) option for authorization.
        LUKS container supports up to 8 keyslots. Parameter value is a string with the new passphrase.
      - NOTE that adding additional passphrase is idempotent only since community.crypto 1.4.0. For older versions, a new
        keyslot will be used even if another keyslot already exists for this passphrase.
      - B(Note) that the passphrase must be UTF-8 encoded text. If you want to use arbitrary binary data, or text using
        another encoding, use the O(passphrase_encoding) option and provide the passphrase Base64 encoded.
    type: str
    version_added: '1.0.0'
  new_keyslot:
    description:
      - Adds the additional O(new_keyfile) or O(new_passphrase) to a specific keyslot on the given O(device). Parameter value
        is the number of the keyslot.
      - B(Note) that a device of O(type=luks1) supports the keyslot numbers V(0)-V(7) and a device of O(type=luks2) supports
        the keyslot numbers V(0)-V(31).
    type: int
    version_added: '2.16.0'
  new_tpm2:
    description:
      - Adds a TPM2 security chip to given container on O(device). Expects a device node path referring to the TPM2 chip (e.g. V(/dev/tpmrm0)).
        Alternatively the special value V(auto) may be specified, in order to automatically determine the device node of a currently
        discovered TPM2 device (of which there must be exactly one). Requires O(new_tpm2_pcrs).
      - B(Note) that O(new_keyslot) does not affect the keyslot for TPM2 enrollment.
      - B(Note) that only LUKS2 containers are supported.
      - B(Note) that systemd-cryptsetup (v248 or newer) is required.
    type: str
    version_added: '3.1.0'
  new_tpm2_pcrs:
    description:
      - TPM2 PCRs (Platform Configuration Registers) to bind to. See systemd-cryptenroll documentation for details (C(--tpm2-pcrs) argument).
    type: str
    version_added: '3.1.0'
  remove_keyfile:
    description:
      - Removes given key from the container on O(device). Does not remove the keyfile from filesystem. Parameter value is
        the path to the keyfile with the passphrase.
      - NOTE that removing keys is idempotent only since community.crypto 1.4.0. For older versions, trying to remove a key
        which no longer exists results in an error.
      - NOTE that to remove the last key from a LUKS container, the O(force_remove_last_key) option must be set to V(true).
      - BEWARE that working with keyfiles in plaintext is dangerous. Make sure that they are protected.
    type: path
  remove_passphrase:
    description:
      - Removes given passphrase from the container on O(device). Parameter value is a string with the passphrase to remove.
      - NOTE that removing passphrases is idempotent only since community.crypto 1.4.0. For older versions, trying to remove
        a passphrase which no longer exists results in an error.
      - NOTE that to remove the last keyslot from a LUKS container, the O(force_remove_last_key) option must be set to V(true).
      - B(Note) that the passphrase must be UTF-8 encoded text. If you want to use arbitrary binary data, or text using
        another encoding, use the O(passphrase_encoding) option and provide the passphrase Base64 encoded.
    type: str
    version_added: '1.0.0'
  remove_tpm2:
    description:
      - Removes B(all) key slots on O(device) that are unlocked by a TPM2 device.
        Needs O(keyfile), O(passphrase), or O(tpm2_device) for authorization.
      - B(Note) that systemd-cryptsetup (v248 or newer) is required.
      - B(Note) that you should avoid using O(tpm2_device) to authorize removal of all TPM2 slots to ensure that you can still access the container afterwards.
    type: bool
    default: false
    version_added: '3.1.0'
  remove_keyslot:
    description:
      - Removes the key in the given slot on O(device). Needs O(keyfile) or O(passphrase) for authorization.
      - B(Note) that a device of O(type=luks1) supports the keyslot numbers V(0)-V(7) and a device of O(type=luks2) supports
        the keyslot numbers V(0)-V(31).
      - B(Note) that the given O(keyfile) or O(passphrase) must not be in the slot to be removed.
    type: int
    version_added: '2.16.0'
  keyslot_priority:
    description:
      - Sets the keyslot priority for the keyslot specified by O(keyslot).
      - B(Note) that keyslot priority is only supported for LUKS2 containers.
    type: str
    choices: [prefer, normal, ignore]
    version_added: '3.1.0'
  force_remove_last_key:
    description:
      - If set to V(true), allows removing the last key from a container.
      - BEWARE that when the last key has been removed from a container, the container can no longer be opened!
    type: bool
    default: false
  label:
    description:
      - This option allow the user to create a LUKS2 format container with label support, respectively to identify the container
        by label on later usages.
      - Will only be used on container creation, or when O(device) is not specified.
      - This cannot be specified if O(type) is set to V(luks1).
    type: str
    version_added: '1.0.0'
  uuid:
    description:
      - With this option user can identify the LUKS container by UUID.
      - Will only be used when O(device) and O(label) are not specified.
    type: str
    version_added: '1.0.0'
  type:
    description:
      - This option allow the user explicit define the format of LUKS container that wants to work with. Options are V(luks1)
        or V(luks2).
    type: str
    choices: [luks1, luks2]
    version_added: '1.0.0'
  cipher:
    description:
      - This option allows the user to define the cipher specification string for the LUKS container.
      - Will only be used on container creation.
      - For pre-2.6.10 kernels, use V(aes-plain) as they do not understand the new cipher spec strings. To use ESSIV, use
        V(aes-cbc-essiv:sha256).
    type: str
    version_added: '1.1.0'
  hash:
    description:
      - This option allows the user to specify the hash function used in LUKS key setup scheme and volume key digest.
      - Will only be used on container creation.
    type: str
    version_added: '1.1.0'
  pbkdf:
    description:
      - This option allows the user to configure the Password-Based Key Derivation Function (PBKDF) used.
      - Will only be used on container creation, and when adding keys to an existing container.
    type: dict
    version_added: '1.4.0'
    suboptions:
      iteration_time:
        description:
          - Specify the iteration time used for the PBKDF.
          - Note that this is in B(seconds), not in milliseconds as on the command line.
          - Mutually exclusive with O(pbkdf.iteration_count).
        type: float
      iteration_count:
        description:
          - Specify the iteration count used for the PBKDF.
          - Mutually exclusive with O(pbkdf.iteration_time).
        type: int
      algorithm:
        description:
          - The algorithm to use.
          - Only available for the LUKS 2 format.
        choices:
          - argon2i
          - argon2id
          - pbkdf2
        type: str
      memory:
        description:
          - The memory cost limit in kilobytes for the PBKDF.
          - This is not used for PBKDF2, but only for the Argon PBKDFs.
        type: int
      parallel:
        description:
          - The parallel cost for the PBKDF. This is the number of threads that run in parallel.
          - This is not used for PBKDF2, but only for the Argon PBKDFs.
        type: int
  sector_size:
    description:
      - This option allows the user to specify the sector size (in bytes) used for LUKS2 containers.
      - Will only be used on container creation.
    type: int
    version_added: '1.5.0'
  perf_same_cpu_crypt:
    description:
      - Allows the user to perform encryption using the same CPU that IO was submitted on.
      - The default is to use an unbound workqueue so that encryption work is automatically balanced between available CPUs.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.3.0'
  perf_submit_from_crypt_cpus:
    description:
      - Allows the user to disable offloading writes to a separate thread after encryption.
      - There are some situations where offloading block write IO operations from the encryption threads to a single thread
        degrades performance significantly.
      - The default is to offload block write IO operations to the same thread.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.3.0'
  perf_no_read_workqueue:
    description:
      - Allows the user to bypass dm-crypt internal workqueue and process read requests synchronously.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.3.0'
  perf_no_write_workqueue:
    description:
      - Allows the user to bypass dm-crypt internal workqueue and process write requests synchronously.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.3.0'
  persistent:
    description:
      - Allows the user to store options into container's metadata persistently and automatically use them next time. Only
        O(perf_same_cpu_crypt), O(perf_submit_from_crypt_cpus), O(perf_no_read_workqueue), O(perf_no_write_workqueue), and
        O(allow_discards) can be stored persistently.
      - Will only work with LUKS2 containers.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.3.0'
  allow_discards:
    description:
      - Allow discards (also known as TRIM) requests for device.
      - Will only be used when opening containers.
    type: bool
    default: false
    version_added: '2.17.0'

requirements:
  - "cryptsetup"
  - "wipefs (when O(state) is V(absent))"
  - "lsblk"
  - "blkid (when O(label) or O(uuid) options are used)"
  - "systemd-cryptsetup (for TPM2 only)"

author: Jan Pokorny (@japokorn)
"""

EXAMPLES = r"""
---
- name: Create LUKS container (remains unchanged if it already exists)
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    keyfile: "/vault/keyfile"

- name: Create LUKS container with a passphrase
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    passphrase: "foo"

- name: Create LUKS container with specific encryption
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    cipher: "aes"
    hash: "sha256"

- name: (Create and) open the LUKS container; name it "mycrypt"
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "opened"
    name: "mycrypt"
    keyfile: "/vault/keyfile"

- name: Close the existing LUKS container "mycrypt"
  community.crypto.luks_device:
    state: "closed"
    name: "mycrypt"

- name: Make sure LUKS container exists and is closed
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "closed"
    keyfile: "/vault/keyfile"

- name: Create container if it does not exist and add new key to it
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    keyfile: "/vault/keyfile"
    new_keyfile: "/vault/keyfile2"

- name: Add new key to the LUKS container (container has to exist)
  community.crypto.luks_device:
    device: "/dev/loop0"
    keyfile: "/vault/keyfile"
    new_keyfile: "/vault/keyfile2"

- name: Add new passphrase to the LUKS container
  community.crypto.luks_device:
    device: "/dev/loop0"
    keyfile: "/vault/keyfile"
    new_passphrase: "foo"

- name: Remove existing keyfile from the LUKS container
  community.crypto.luks_device:
    device: "/dev/loop0"
    remove_keyfile: "/vault/keyfile2"

- name: Remove existing passphrase from the LUKS container
  community.crypto.luks_device:
    device: "/dev/loop0"
    remove_passphrase: "foo"

- name: Completely remove the LUKS container and its contents
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "absent"

- name: Create a container with label
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    keyfile: "/vault/keyfile"
    label: personalLabelName

- name: Open the LUKS container based on label without device; name it "mycrypt"
  community.crypto.luks_device:
    label: "personalLabelName"
    state: "opened"
    name: "mycrypt"
    keyfile: "/vault/keyfile"

- name: Close container based on UUID
  community.crypto.luks_device:
    uuid: 03ecd578-fad4-4e6c-9348-842e3e8fa340
    state: "closed"
    name: "mycrypt"

- name: Create a container using luks2 format
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    keyfile: "/vault/keyfile"
    type: luks2

- name: Create a container with key in slot 4
  community.crypto.luks_device:
    device: "/dev/loop0"
    state: "present"
    keyfile: "/vault/keyfile"
    keyslot: 4

- name: Add a new key in slot 5
  community.crypto.luks_device:
    device: "/dev/loop0"
    keyfile: "/vault/keyfile"
    new_keyfile: "/vault/keyfile"
    new_keyslot: 5

- name: Remove the key from slot 4 (given keyfile must not be slot 4)
  community.crypto.luks_device:
    device: "/dev/loop0"
    keyfile: "/vault/keyfile"
    remove_keyslot: 4

- name: Enroll a TPM2 device using a keyfile to unlock the container
  community.crypto.luks_device:
    keyfile: "/vault/keyfile"
    new_tpm2: "auto"
    new_tpm2_pcrs: "1+3+5+7+11+12+14"

- name: Remove all enrolled TPM2 devices
  community.crypto.luks_device:
    tpm2_device: "auto"
    remove_tpm2: true

- name: Set the priority of keyslot 0 to 'prefer'
  community.crypto.luks_device:
    device: "/dev/loop0"
    keyslot: 0
    keyslot_priority: prefer
"""

RETURN = r"""
name:
  description: When O(state=opened) returns (generated or given) name of LUKS container. Returns None if no name is supplied.
  returned: success
  type: str
  sample: "luks-c1da9a58-2fde-4256-9d9f-6ab008b4dd1b"
"""

import json
import os
import re
import stat
import typing as t
from base64 import b64decode

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text

# used to get <luks-name> out of lsblk output in format 'crypt <luks-name>'
# regex takes care of any possible blank characters
LUKS_NAME_REGEX = re.compile(r"^crypt\s+([^\s]*)\s*$")
# used to get </luks/device> out of lsblk output
# in format 'device: </luks/device>'
LUKS_DEVICE_REGEX = re.compile(r"\s*device:\s+([^\s]*)\s*")


# See https://gitlab.com/cryptsetup/cryptsetup/-/wikis/LUKS-standard/on-disk-format.pdf
LUKS_HEADER = b"LUKS\xba\xbe"
LUKS_HEADER_L = 6
# See https://gitlab.com/cryptsetup/LUKS2-docs/-/blob/master/luks2_doc_wip.pdf
LUKS2_HEADER_OFFSETS = [
    0x4000,
    0x8000,
    0x10000,
    0x20000,
    0x40000,
    0x80000,
    0x100000,
    0x200000,
    0x400000,
]
LUKS2_HEADER2 = b"SKUL\xba\xbe"


def wipe_luks_headers(device: str) -> None:
    wipe_offsets = []
    with open(device, "rb") as f:
        # f.seek(0)
        data = f.read(LUKS_HEADER_L)
        if data == LUKS_HEADER:
            wipe_offsets.append(0)
        for offset in LUKS2_HEADER_OFFSETS:
            f.seek(offset)
            data = f.read(LUKS_HEADER_L)
            if data == LUKS2_HEADER2:
                wipe_offsets.append(offset)

    if wipe_offsets:
        with open(device, "wb") as f:
            for offset in wipe_offsets:
                f.seek(offset)
                f.write(b"\x00\x00\x00\x00\x00\x00")


class Handler:
    def __init__(self, module: AnsibleModule) -> None:
        self._module = module
        self._lsblk_bin = self._module.get_bin_path("lsblk", True)
        self._passphrase_encoding = module.params["passphrase_encoding"]

    def get_passphrase_from_module_params(self, parameter_name: str) -> bytes | None:
        passphrase = self._module.params[parameter_name]
        if passphrase is None:
            return None
        if self._passphrase_encoding == "text":
            return to_bytes(passphrase)
        try:
            return b64decode(to_text(passphrase))
        except Exception as exc:
            self._module.fail_json(
                f"Error while base64-decoding '{parameter_name}': {exc}"
            )

    def _run_command(
        self, command: list[str], data: bytes | None = None
    ) -> tuple[int, str, str]:
        return self._module.run_command(command, data=data, binary_data=True)

    def get_device_by_uuid(self, uuid: str | None) -> str | None:
        """Returns the device that holds UUID passed by user"""
        blkid_bin = self._module.get_bin_path("blkid", True)
        if uuid is None:
            return None
        rc, stdout, dummy = self._run_command([blkid_bin, "--uuid", uuid])
        if rc != 0:
            return None
        return stdout.strip()

    def get_device_by_label(self, label: str) -> str | None:
        """Returns the device that holds label passed by user"""
        blkid_bin = self._module.get_bin_path("blkid", True)
        rc, stdout, dummy = self._run_command([blkid_bin, "--label", label])
        if rc != 0:
            return None
        return stdout.strip()

    def generate_luks_name(self, device: str) -> str:
        """Generate name for luks based on device UUID ('luks-<UUID>').
        Raises ValueError when obtaining of UUID fails.
        """
        rc, stdout, stderr = self._run_command(
            [self._lsblk_bin, "-n", device, "-o", "UUID"]
        )

        if rc != 0:
            raise ValueError(f"Error while generating LUKS name for {device}: {stderr}")
        dev_uuid = stdout.strip()
        return f"luks-{dev_uuid}"


class CryptHandler(Handler):
    def __init__(self, module: AnsibleModule) -> None:
        super().__init__(module)
        self._cryptsetup_bin = self._module.get_bin_path("cryptsetup", True)

    def get_container_name_by_device(self, device: str) -> str | None:
        """obtain LUKS container name based on the device where it is located
        return None if not found
        raise ValueError if lsblk command fails
        """
        rc, stdout, stderr = self._run_command(
            [self._lsblk_bin, device, "-nlo", "type,name"]
        )
        if rc != 0:
            raise ValueError(f"Error while obtaining LUKS name for {device}: {stderr}")

        for line in stdout.splitlines(False):
            m = LUKS_NAME_REGEX.match(line)
            if m:
                return m.group(1)
        return None

    def get_container_device_by_name(self, name: str) -> str | None:
        """obtain device name based on the LUKS container name
        return None if not found
        raise ValueError if lsblk command fails
        """
        # apparently each device can have only one LUKS container on it
        rc, stdout, dummy = self._run_command([self._cryptsetup_bin, "status", name])
        if rc != 0:
            return None

        m = LUKS_DEVICE_REGEX.search(stdout)
        if not m:
            return None
        device = m.group(1)
        return device

    def is_luks(self, device: str) -> bool:
        """check if the LUKS container does exist"""
        rc, dummy, dummy2 = self._run_command([self._cryptsetup_bin, "isLuks", device])
        return rc == 0

    def get_luks_type(self, device: str) -> t.Literal["luks1", "luks2"] | None:
        """get the luks type of a device"""
        if self.is_luks(device):
            with open(device, "rb") as f:
                for offset in LUKS2_HEADER_OFFSETS:
                    f.seek(offset)
                    data = f.read(LUKS_HEADER_L)
                    if data == LUKS2_HEADER2:
                        return "luks2"
                return "luks1"
        return None

    def is_luks_slot_set(self, device: str, keyslot: int) -> bool:
        """check if a keyslot is set"""
        rc, stdout, dummy = self._run_command(
            [self._cryptsetup_bin, "luksDump", device]
        )
        if rc != 0:
            raise ValueError(f"Error while dumping LUKS header from {device}")
        result_luks1 = f"Key Slot {keyslot}: ENABLED" in stdout
        result_luks2 = f" {keyslot}: luks2" in stdout
        return result_luks1 or result_luks2

    def _add_pbkdf_options(self, options: list[str], pbkdf: dict[str, t.Any]) -> None:
        if pbkdf["iteration_time"] is not None:
            options.extend(["--iter-time", str(int(pbkdf["iteration_time"] * 1000))])
        if pbkdf["iteration_count"] is not None:
            options.extend(["--pbkdf-force-iterations", str(pbkdf["iteration_count"])])
        if pbkdf["algorithm"] is not None:
            options.extend(["--pbkdf", pbkdf["algorithm"]])
        if pbkdf["memory"] is not None:
            options.extend(["--pbkdf-memory", str(pbkdf["memory"])])
        if pbkdf["parallel"] is not None:
            options.extend(["--pbkdf-parallel", str(pbkdf["parallel"])])

    def run_luks_create(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        keyslot: int | None,
        keysize: int | None,
        cipher: str | None,
        hash_: str | None,
        sector_size: str | None,
        pbkdf: dict[str, t.Any] | None,
    ) -> None:
        # create a new luks container; use batch mode to auto confirm
        luks_type = self._module.params["type"]
        label = self._module.params["label"]

        options = []
        if keysize is not None:
            options.append("--key-size=" + str(keysize))
        if label is not None:
            options.extend(["--label", label])
            luks_type = "luks2"
        if luks_type is not None:
            options.extend(["--type", luks_type])
        if cipher is not None:
            options.extend(["--cipher", cipher])
        if hash_ is not None:
            options.extend(["--hash", hash_])
        if pbkdf is not None:
            self._add_pbkdf_options(options, pbkdf)
        if sector_size is not None:
            options.extend(["--sector-size", str(sector_size)])
        if keyslot is not None:
            options.extend(["--key-slot", str(keyslot)])

        args = [self._cryptsetup_bin, "luksFormat"]
        args.extend(options)
        args.extend(["-q", device])
        if keyfile:
            args.append(keyfile)
        else:
            args.append("-")

        rc, dummy, stderr = self._run_command(args, data=passphrase)
        if rc != 0:
            raise ValueError(f"Error while creating LUKS on {device}: {stderr}")

    def run_luks_open(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        perf_same_cpu_crypt: bool,
        perf_submit_from_crypt_cpus: bool,
        perf_no_read_workqueue: bool,
        perf_no_write_workqueue: bool,
        persistent: bool,
        allow_discards: bool,
        name: str,
    ) -> None:
        args = [self._cryptsetup_bin]
        if keyfile:
            args.extend(["--key-file", keyfile])
        else:
            args.extend(["--key-file", "-"])
        if perf_same_cpu_crypt:
            args.extend(["--perf-same_cpu_crypt"])
        if perf_submit_from_crypt_cpus:
            args.extend(["--perf-submit_from_crypt_cpus"])
        if perf_no_read_workqueue:
            args.extend(["--perf-no_read_workqueue"])
        if perf_no_write_workqueue:
            args.extend(["--perf-no_write_workqueue"])
        if persistent:
            args.extend(["--persistent"])
        if allow_discards:
            args.extend(["--allow-discards"])
        args.extend(["open", "--type", "luks", device, name])

        rc, dummy, stderr = self._run_command(args, data=passphrase)
        if rc != 0:
            raise ValueError(
                f"Error while opening LUKS container on {device}: {stderr}"
            )

    def run_luks_close(self, name: str) -> None:
        rc, dummy, dummy2 = self._run_command([self._cryptsetup_bin, "close", name])
        if rc != 0:
            raise ValueError(f"Error while closing LUKS container {name}")

    def run_luks_remove(self, device: str) -> None:
        wipefs_bin = self._module.get_bin_path("wipefs", True)

        name = self.get_container_name_by_device(device)
        if name is not None:
            self.run_luks_close(name)
        rc, dummy, stderr = self._run_command([wipefs_bin, "--all", device])
        if rc != 0:
            raise ValueError(
                f"Error while wiping LUKS container signatures for {device}: {stderr}"
            )

        # For LUKS2, sometimes both `cryptsetup erase` and `wipefs` do **not**
        # erase all LUKS signatures (they seem to miss the second header). That's
        # why we do it ourselves here.
        try:
            wipe_luks_headers(device)
        except Exception as exc:
            raise ValueError(
                f"Error while wiping LUKS container signatures for {device}: {exc}"
            ) from exc

    def run_luks_add_key(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        new_keyfile: str | None,
        new_passphrase: bytes | None,
        new_keyslot: int | None,
        pbkdf: dict[str, t.Any] | None,
    ) -> None:
        """Add new key from a keyfile or passphrase to given 'device';
        authentication done using 'keyfile' or 'passphrase'.
        Raises ValueError when command fails.
        """
        data = []
        args = [self._cryptsetup_bin, "luksAddKey", device]
        if pbkdf is not None:
            self._add_pbkdf_options(args, pbkdf)

        if new_keyslot is not None:
            args.extend(["--key-slot", str(new_keyslot)])

        if keyfile:
            args.extend(["--key-file", keyfile])
        elif passphrase is not None:
            args.extend(["--key-file", "-", "--keyfile-size", str(len(passphrase))])
            data.append(passphrase)
        else:
            raise ValueError("Need passphrase or keyfile")

        if new_keyfile:
            args.append(new_keyfile)
        elif new_passphrase is not None:
            args.append("-")
            data.append(new_passphrase)
        else:
            raise ValueError("Need new passphrase or new keyfile")

        rc, dummy, stderr = self._run_command(args, data=b"".join(data) or None)
        if rc != 0:
            raise ValueError(
                f"Error while adding new LUKS keyslot to {device}: {stderr}"
            )

    def run_luks_remove_key(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        keyslot: int | None,
        force_remove_last_key: bool = False,
    ) -> None:
        """Remove key from given device
        Raises ValueError when command fails
        """
        if not force_remove_last_key:
            rc, stdout, dummy = self._run_command(
                [self._cryptsetup_bin, "luksDump", device]
            )
            if rc != 0:
                raise ValueError(f"Error while dumping LUKS header from {device}")
            keyslot_count = 0
            keyslot_area = False
            keyslot_re = re.compile(r"^Key Slot [0-9]+: ENABLED")
            for line in stdout.splitlines():
                if line.startswith("Keyslots:"):
                    keyslot_area = True
                elif line.startswith("  "):
                    # LUKS2 header dumps use human-readable indented output.
                    # Thus we have to look out for 'Keyslots:' and count the
                    # number of indented keyslot numbers.
                    if keyslot_area and line[2] in "0123456789":
                        keyslot_count += 1
                elif line.startswith("\t"):
                    pass
                elif keyslot_re.match(line):
                    # LUKS1 header dumps have one line per keyslot with ENABLED
                    # or DISABLED in them. We count such lines with ENABLED.
                    keyslot_count += 1
                else:
                    keyslot_area = False
            if keyslot_count < 2:
                self._module.fail_json(
                    msg=f"LUKS device {device} has less than two active keyslots. To be able to remove a key, please set `force_remove_last_key` to `true`."
                )

        if keyslot is None:
            args = [self._cryptsetup_bin, "luksRemoveKey", device, "-q"]
            if keyfile:
                args.extend(["--key-file", keyfile])
            elif passphrase is not None:
                args.extend(["--key-file", "-"])
        else:
            # Since we supply -q no passphrase is needed
            args = [self._cryptsetup_bin, "luksKillSlot", device, "-q", str(keyslot)]
            passphrase = None
        rc, dummy, stderr = self._run_command(args, data=passphrase)
        if rc != 0:
            raise ValueError(f"Error while removing LUKS key from {device}: {stderr}")

    def luks_test_key(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        keyslot: int | None = None,
    ) -> bool:
        """Check whether the keyfile or passphrase works.
        Raises ValueError when command fails.
        """
        data = None
        args = [self._cryptsetup_bin, "luksOpen", "--test-passphrase", device]

        if keyfile:
            args.extend(["--key-file", keyfile])
        else:
            args.extend(["--key-file", "-"])
            data = passphrase

        if keyslot is not None:
            args.extend(["--key-slot", str(keyslot)])

        rc, stdout, stderr = self._run_command(args, data=data)
        if rc == 0:
            return True
        for output in (stdout, stderr):
            if "No key available with this passphrase" in output:
                return False
            if "No usable keyslot is available." in output:
                return False

        # This check is necessary due to cryptsetup in version 2.0.3 not printing 'No usable keyslot is available'
        # when using the --key-slot parameter in combination with --test-passphrase
        if rc == 1 and keyslot is not None and stdout == "" and stderr == "":
            return False

        raise ValueError(
            f"Error while testing whether keyslot exists on {device}: {stderr}"
        )

    def luks_dump_json_metadata(self, device: str) -> dict:
        """Dump LUKS metadata in JSON format.
        Raises ValueError when command fails.
        """
        rc, stdout, stderr = self._run_command(
            [self._cryptsetup_bin, "luksDump", "--dump-json-metadata", device]
        )
        if rc != 0:
            raise ValueError(
                f"Error while dumping LUKS JSON metadata from {device}: {stderr}"
            )

        return json.loads(stdout)

    def run_config(self, device: str, keyslot: int, keyslot_priority: str) -> None:
        """Configure LUKS keyslot priority.
        Raises ValueError when command fails.
        """
        rc, dummy, stderr = self._run_command(
            [
                self._cryptsetup_bin,
                "config",
                f"--key-slot={keyslot}",
                f"--priority={keyslot_priority}",
                device,
            ]
        )
        if rc != 0:
            raise ValueError(
                f"Error while configuring LUKS keyslot {keyslot} on {device}: {stderr}"
            )

    def run_systemd_cryptsetup_attach(
        self,
        device: str,
        tpm2_device: str | None,
        perf_same_cpu_crypt: bool,
        perf_submit_from_crypt_cpus: bool,
        perf_no_read_workqueue: bool,
        perf_no_write_workqueue: bool,
        allow_discards: bool,
        name: str,
    ) -> None:
        systemd_cryptsetup_bin = self._module.get_bin_path(
            "systemd-cryptsetup", required=True
        )
        args = [systemd_cryptsetup_bin, name, device, "none"]
        options = []

        if tpm2_device is not None:
            options.append(f"tpm2-device={tpm2_device}")

        if perf_same_cpu_crypt:
            options.append("same-cpu-crypt")

        if perf_submit_from_crypt_cpus:
            options.append("submit-from-crypt-cpus")

        if perf_no_read_workqueue:
            options.append("no-read-workqueue")

        if perf_no_write_workqueue:
            options.append("no-write-workqueue")

        if allow_discards:
            options.append("discard")

        args.append(",".join(options))

        rc, dummy, stderr = self._run_command(args)

        if rc != 0:
            raise ValueError(f"Error while opening {device}: {stderr}")

    def run_systemd_cryptenroll(
        self,
        device: str,
        keyfile: str | None,
        passphrase: bytes | None,
        tpm2_device: str | None,
        new_tpm2: str | None,
        new_tpm2_pcrs: str | None,
        remove_tpm2: bool,
    ) -> bool:
        systemd_cryptenroll_bin = self._module.get_bin_path(
            "systemd-cryptenroll", required=True
        )
        args = [systemd_cryptenroll_bin]

        if keyfile is not None:
            args.append(f"--unlock-key-file={keyfile}")

        if tpm2_device:
            args.append(f"--unlock-tpm2-device={tpm2_device}")

        if new_tpm2:
            args.extend([f"--tpm2-device={new_tpm2}", f"--tpm2-pcrs={new_tpm2_pcrs}"])

        if remove_tpm2:
            args.append("--wipe-slot=tpm2")

        if passphrase:
            # --unlock-key-file is needed both so a newline isnt required at the end of stdin
            # and also because otherwise, systemd-cryptenroll will just hang if the passphrase is wrong,
            # waiting for the user to input another one
            rc, dummy, stderr = self._run_command(
                [*args, "--unlock-key-file=/dev/stdin", device], data=passphrase
            )
        else:
            rc, dummy, stderr = self._run_command([*args, device])

        if rc != 0:
            raise ValueError(
                f"Error while adding key to {device} with {args}: {stderr}"
            )

        # systemd-cryptenroll stores a policy hash in the LUKS metadata which is used to
        # detect if a TPM2 device and PCR set are already enrolled. Computing it
        # is very complicated, however, so rather than duplicating all of that here,
        # we rely on the command's output which hopefully won't change.
        # See: https://github.com/systemd/systemd/blob/7524671f74c9b0ea858a077ae9b1af3fe574d57e/src/cryptenroll/cryptenroll-tpm2.c#L529-L541
        return "This PCR set is already enrolled, executing no operation." not in stderr


class ConditionsHandler(Handler):
    def __init__(self, module: AnsibleModule, crypthandler: CryptHandler) -> None:
        super().__init__(module)
        self._crypthandler = crypthandler
        self.device = self.get_device_name()

    def get_device_name(self) -> str | None:
        device: str | None = self._module.params.get("device")
        label: str | None = self._module.params.get("label")
        uuid: str | None = self._module.params.get("uuid")
        name: str | None = self._module.params.get("name")

        if device is None and label is not None:
            device = self.get_device_by_label(label)
        elif device is None and uuid is not None:
            device = self.get_device_by_uuid(uuid)
        elif device is None and name is not None:
            device = self._crypthandler.get_container_device_by_name(name)

        return device

    def luks_create(self) -> bool:
        if (
            self.device is not None
            and self._module.params["state"] in ("present", "opened", "closed")
            and not self._crypthandler.is_luks(self.device)
        ):

            if (
                self._module.params["keyfile"] is None
                and self._module.params["passphrase"] is None
            ):

                self._module.fail_json(
                    msg="Neither keyfile nor passphrase were given but the LUKS volume must be created"
                )

            return True

        return False

    def opened_luks_name(self, device: str) -> str | None:
        """If luks is already opened, return its name.
        If 'name' parameter is specified and differs
        from obtained value, fail.
        Return None otherwise
        """
        if self._module.params["state"] != "opened":
            return None

        # try to obtain luks name - it may be already opened
        name = self._crypthandler.get_container_name_by_device(device)

        if name is None:
            # container is not open
            return None

        if self._module.params["name"] is None:
            # container is already opened
            return name

        if name != self._module.params["name"]:
            # the container is already open but with different name:
            # suspicious. back off
            self._module.fail_json(
                msg=f"LUKS container is already opened under different name '{name}'."
            )

        # container is opened and the names match
        return name

    def luks_open(self) -> bool:
        if self.device is None or self._module.params["state"] != "opened":
            # conditions for open not fulfilled
            return False

        if (
            self._module.params["keyfile"] is None
            and self._module.params["passphrase"] is None
            and self._module.params["tpm2_device"] is None
        ):
            self._module.fail_json(
                msg="state=opened was specified but none of keyfile, passphrase, or tpm2_device were given"
            )

        name = self.opened_luks_name(self.device)

        return name is None

    def luks_close(self) -> bool:
        if (
            self._module.params["name"] is None and self.device is None
        ) or self._module.params["state"] != "closed":
            # conditions for close not fulfilled
            return False
        luks_is_open = False

        if self.device is not None:
            name = self._crypthandler.get_container_name_by_device(self.device)
            # successfully getting name based on device means that luks is open
            luks_is_open = name is not None

        if self._module.params["name"] is not None:
            device = self._crypthandler.get_container_device_by_name(
                self._module.params["name"]
            )
            # successfully getting device based on name means that luks is open
            luks_is_open = device is not None
            if device is not None:
                self.device = device

        return luks_is_open

    def luks_add_key(self) -> bool:
        if (
            self.device is None
            or (
                self._module.params["keyfile"] is None
                and self._module.params["passphrase"] is None
            )
            or (
                self._module.params["new_keyfile"] is None
                and self._module.params["new_passphrase"] is None
            )
        ):
            # conditions for adding a key not fulfilled
            return False

        if self._module.params["state"] == "absent":
            self._module.fail_json(
                msg="Contradiction in setup: Asking to add a key to absent LUKS."
            )

        key_present = self._crypthandler.luks_test_key(
            self.device,
            self._module.params["new_keyfile"],
            self.get_passphrase_from_module_params("new_passphrase"),
        )
        if self._module.params["new_keyslot"] is not None:
            key_present_slot = self._crypthandler.luks_test_key(
                self.device,
                self._module.params["new_keyfile"],
                self.get_passphrase_from_module_params("new_passphrase"),
                self._module.params["new_keyslot"],
            )
            if key_present and not key_present_slot:
                self._module.fail_json(
                    msg="Trying to add key that is already present in another slot"
                )

        return not key_present

    def luks_remove_key(self) -> bool:
        if self.device is None or (
            self._module.params["remove_keyfile"] is None
            and self._module.params["remove_passphrase"] is None
            and self._module.params["remove_keyslot"] is None
        ):
            # conditions for removing a key not fulfilled
            return False

        if self._module.params["state"] == "absent":
            self._module.fail_json(
                msg="Contradiction in setup: Asking to remove a key from absent LUKS."
            )

        if self._module.params["remove_keyslot"] is not None:
            if not self._crypthandler.is_luks_slot_set(
                self.device, self._module.params["remove_keyslot"]
            ):
                return False
            result = self._crypthandler.luks_test_key(
                self.device,
                self._module.params["keyfile"],
                self.get_passphrase_from_module_params("passphrase"),
            )
            if self._crypthandler.luks_test_key(
                self.device,
                self._module.params["keyfile"],
                self.get_passphrase_from_module_params("passphrase"),
                self._module.params["remove_keyslot"],
            ):
                self._module.fail_json(
                    msg="Cannot remove keyslot with keyfile or passphrase in same slot."
                )
            return result

        return self._crypthandler.luks_test_key(
            self.device,
            self._module.params["remove_keyfile"],
            self.get_passphrase_from_module_params("remove_passphrase"),
        )

    def luks_remove(self) -> bool:
        return (
            self.device is not None
            and self._module.params["state"] == "absent"
            and self._crypthandler.is_luks(self.device)
        )

    def luks_config(self) -> bool:
        if self.device is None or self._module.params["keyslot_priority"] is None:
            return False

        if self._module.params["keyslot"] is None:
            self._module.fail_json(
                msg="keyslot_priority was specified but keyslot was not."
            )

        json_metadata = self._crypthandler.luks_dump_json_metadata(self.device)
        slot_priority = (
            json_metadata.get("keyslots", {})
            .get(str(self._module.params["keyslot"]), {})
            .get("priority", None)
        )

        priority_map = {"prefer": 2, "normal": None, "ignore": 0}
        return slot_priority != priority_map[self._module.params["keyslot_priority"]]

    def validate_keyslot(
        self, param: str, luks_type: t.Literal["luks1", "luks2"] | None
    ) -> None:
        if self._module.params[param] is not None:
            if luks_type is None and param == "keyslot":
                if 8 <= self._module.params[param] <= 31:
                    self._module.fail_json(
                        msg="You must specify type=luks2 when creating a new LUKS device to use keyslots 8-31."
                    )
                elif not (0 <= self._module.params[param] <= 7):
                    self._module.fail_json(
                        msg="When not specifying a type, only the keyslots 0-7 are allowed."
                    )

            if luks_type == "luks1" and not 0 <= self._module.params[param] <= 7:
                self._module.fail_json(
                    msg=f"{self._module.params[param]} must be between 0 and 7 when using LUKS1."
                )
            elif luks_type == "luks2" and not 0 <= self._module.params[param] <= 31:
                self._module.fail_json(
                    msg=f"{self._module.params[param]} must be between 0 and 31 when using LUKS2."
                )

    def systemd_cryptenroll(self) -> bool:
        if self.device is None:
            return False

        if (
            self._module.params["new_tpm2"] is not None
            and self._module.params["new_tpm2_pcrs"] is None
        ):
            self._module.fail_json(msg="new_tpm2_pcrs must be specified with new_tpm2")

        if self._module.params["remove_tpm2"]:
            json_metadata = self._crypthandler.luks_dump_json_metadata(self.device)
            tokens = json_metadata.get("tokens", {})
            tpm2_enrolled = any(
                token.get("type") == "systemd-tpm2" for token in tokens.values()
            )
            return tpm2_enrolled or self._module.params["new_tpm2"] is not None

        return self._module.params["new_tpm2"] is not None


def run_module() -> t.NoReturn:  # noqa: C901
    # available arguments/parameters that a user can pass
    module_args = {
        "state": {
            "type": "str",
            "default": "present",
            "choices": ["present", "absent", "opened", "closed"],
        },
        "device": {"type": "str"},
        "name": {"type": "str"},
        "keyfile": {"type": "path"},
        "new_keyfile": {"type": "path"},
        "remove_keyfile": {"type": "path"},
        "passphrase": {"type": "str", "no_log": True},
        "new_passphrase": {"type": "str", "no_log": True},
        "remove_passphrase": {"type": "str", "no_log": True},
        "tpm2_device": {"type": "str"},
        "new_tpm2": {"type": "str"},
        "new_tpm2_pcrs": {"type": "str"},
        "remove_tpm2": {"type": "bool", "default": False},
        "passphrase_encoding": {
            "type": "str",
            "default": "text",
            "choices": ["text", "base64"],
            "no_log": False,
        },
        "keyslot": {"type": "int", "no_log": False},
        "keyslot_priority": {
            "type": "str",
            "no_log": False,
            "choices": ["prefer", "normal", "ignore"],
        },
        "new_keyslot": {"type": "int", "no_log": False},
        "remove_keyslot": {"type": "int", "no_log": False},
        "force_remove_last_key": {"type": "bool", "default": False},
        "keysize": {"type": "int"},
        "label": {"type": "str"},
        "uuid": {"type": "str"},
        "type": {"type": "str", "choices": ["luks1", "luks2"]},
        "cipher": {"type": "str"},
        "hash": {"type": "str"},
        "pbkdf": {
            "type": "dict",
            "options": {
                "iteration_time": {"type": "float"},
                "iteration_count": {"type": "int"},
                "algorithm": {
                    "type": "str",
                    "choices": ["argon2i", "argon2id", "pbkdf2"],
                },
                "memory": {"type": "int"},
                "parallel": {"type": "int"},
            },
            "mutually_exclusive": [("iteration_time", "iteration_count")],
        },
        "sector_size": {"type": "int"},
        "perf_same_cpu_crypt": {"type": "bool", "default": False},
        "perf_submit_from_crypt_cpus": {"type": "bool", "default": False},
        "perf_no_read_workqueue": {"type": "bool", "default": False},
        "perf_no_write_workqueue": {"type": "bool", "default": False},
        "persistent": {"type": "bool", "default": False},
        "allow_discards": {"type": "bool", "default": False},
    }

    mutually_exclusive = [
        ("keyfile", "passphrase", "tpm2_device"),
        ("new_keyfile", "new_passphrase"),
        ("remove_keyfile", "remove_passphrase", "remove_keyslot"),
    ]

    required_by = {
        "new_tpm2": ["new_tpm2_pcrs"],
        "keyslot_priority": ["keyslot"],
    }

    # seed the result dict in the object
    result: dict[str, t.Any] = {"changed": False, "name": None}

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True,
        mutually_exclusive=mutually_exclusive,
        required_by=required_by,
    )
    module.run_command_environ_update = {
        "LANG": "C",
        "LC_ALL": "C",
        "LC_MESSAGES": "C",
        "LC_CTYPE": "C",
    }

    if module.params["device"] is not None:
        try:
            statinfo = os.stat(module.params["device"])
            mode = statinfo.st_mode
            if not stat.S_ISBLK(mode) and not stat.S_ISCHR(mode):
                raise Exception(f"{module.params['device']} is not a device")
        except Exception as e:
            module.fail_json(msg=str(e))

    # conditions not allowed to run
    if module.params["label"] is not None and module.params["type"] == "luks1":
        module.fail_json(msg="You cannot combine type luks1 with the label option.")

    crypt = CryptHandler(module)
    try:
        conditions = ConditionsHandler(module, crypt)
    except ValueError as exc:
        module.fail_json(msg=str(exc))

    if (
        module.params["keyslot"] is not None
        or module.params["new_keyslot"] is not None
        or module.params["remove_keyslot"] is not None
    ):
        luks_type = (
            crypt.get_luks_type(conditions.device)
            if conditions.device is not None
            else None
        )
        if luks_type is None and module.params["type"] is not None:
            luks_type = module.params["type"]
        for param in ["keyslot", "new_keyslot", "remove_keyslot"]:
            conditions.validate_keyslot(param, luks_type)

    for param in ["new_keyslot", "remove_keyslot"]:
        if (
            module.params[param] is not None
            and module.params["keyfile"] is None
            and module.params["passphrase"] is None
        ):
            module.fail_json(
                msg="Removing a keyslot requires the passphrase or keyfile of another slot."
            )

    # The conditions are in order to allow more operations in one run.
    # (e.g. create luks and add a key to it)

    # luks create
    if conditions.luks_create():
        assert conditions.device  # ensured in conditions.luks_create()
        if not module.check_mode:
            try:
                crypt.run_luks_create(
                    conditions.device,
                    module.params["keyfile"],
                    conditions.get_passphrase_from_module_params("passphrase"),
                    module.params["keyslot"],
                    module.params["keysize"],
                    module.params["cipher"],
                    module.params["hash"],
                    module.params["sector_size"],
                    module.params["pbkdf"],
                )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # luks open

    if conditions.device is not None:
        name = conditions.opened_luks_name(conditions.device)
        if name is not None:
            result["name"] = name

    if conditions.luks_open():
        assert conditions.device  # ensured in conditions.luks_open()
        name = module.params["name"]
        if name is None:
            try:
                name = crypt.generate_luks_name(conditions.device)
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        if not module.check_mode:
            try:
                if (
                    module.params["keyfile"] is not None
                    or module.params["passphrase"] is not None
                ):
                    crypt.run_luks_open(
                        conditions.device,
                        module.params["keyfile"],
                        conditions.get_passphrase_from_module_params("passphrase"),
                        module.params["perf_same_cpu_crypt"],
                        module.params["perf_submit_from_crypt_cpus"],
                        module.params["perf_no_read_workqueue"],
                        module.params["perf_no_write_workqueue"],
                        module.params["persistent"],
                        module.params["allow_discards"],
                        name,
                    )
                else:
                    crypt.run_systemd_cryptsetup_attach(
                        conditions.device,
                        module.params["tpm2_device"],
                        module.params["perf_same_cpu_crypt"],
                        module.params["perf_submit_from_crypt_cpus"],
                        module.params["perf_no_read_workqueue"],
                        module.params["perf_no_write_workqueue"],
                        module.params["allow_discards"],
                        name,
                    )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["name"] = name
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # luks close
    if conditions.luks_close():
        if conditions.device is not None:
            try:
                name = crypt.get_container_name_by_device(conditions.device)
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        else:
            name = module.params["name"]
        if name is None:
            module.fail_json(msg="Cannot determine name to close device")
        if not module.check_mode:
            try:
                crypt.run_luks_close(name)
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["name"] = name
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # luks add key
    if conditions.luks_add_key():
        assert conditions.device  # ensured in conditions.luks_add_key()
        if not module.check_mode:
            try:
                crypt.run_luks_add_key(
                    conditions.device,
                    module.params["keyfile"],
                    conditions.get_passphrase_from_module_params("passphrase"),
                    module.params["new_keyfile"],
                    conditions.get_passphrase_from_module_params("new_passphrase"),
                    module.params["new_keyslot"],
                    module.params["pbkdf"],
                )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # luks remove key
    if conditions.luks_remove_key():
        assert conditions.device  # ensured in conditions.luks_remove_key()
        if not module.check_mode:
            try:
                last_key = module.params["force_remove_last_key"]
                crypt.run_luks_remove_key(
                    conditions.device,
                    module.params["remove_keyfile"],
                    conditions.get_passphrase_from_module_params("remove_passphrase"),
                    module.params["remove_keyslot"],
                    force_remove_last_key=last_key,
                )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # luks remove
    if conditions.luks_remove():
        assert conditions.device  # ensured in conditions.luks_remove()
        if not module.check_mode:
            try:
                crypt.run_luks_remove(conditions.device)
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    if conditions.systemd_cryptenroll():
        assert conditions.device is not None
        if not module.check_mode:
            try:
                changed = crypt.run_systemd_cryptenroll(
                    conditions.device,
                    module.params["keyfile"],
                    conditions.get_passphrase_from_module_params("passphrase"),
                    module.params["tpm2_device"],
                    module.params["new_tpm2"],
                    module.params["new_tpm2_pcrs"],
                    module.params["remove_tpm2"],
                )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
            result["changed"] |= changed
        if module.check_mode:
            module.exit_json(**result)

    if conditions.luks_config():
        assert conditions.device is not None  # ensured in conditions.luks_config()
        if not module.check_mode:
            try:
                crypt.run_config(
                    conditions.device,
                    module.params["keyslot"],
                    module.params["keyslot_priority"],
                )
            except ValueError as e:
                module.fail_json(msg=f"luks_device error: {e}")
        result["changed"] = True
        if module.check_mode:
            module.exit_json(**result)

    # Success - return result
    module.exit_json(**result)


def main() -> t.NoReturn:
    run_module()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The ios vrf_address_family fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.vrf_address_family.vrf_address_family import (
    Vrf_address_familyArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.rm_templates.vrf_address_family import (
    Vrf_address_familyTemplate,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.utils.utils import (
    flatten_config,
)


class Vrf_address_familyFacts(object):
    """The ios vrf_address_family facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Vrf_address_familyArgs.argument_spec

    def get_config(self, connection):
        """Get the configuration from the device"""

        return connection.get("show running-config | section ^vrf")

    def _flatten_config(self, config):
        dataLines = config.split("\n")
        finalConfig = []

        for line in dataLines:
            if "address-family" in line and "exit-address-family" not in line:
                finalConfig.append(line)

        return "\n".join(finalConfig)

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for Vrf_address_family network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        objs = []
        obj = {}

        if not data:
            data = self.get_config(connection)

        address_data = flatten_config(data, "address-family")
        data = flatten_config(address_data, "vrf")
        finalConfig = self._flatten_config(data)

        # parse native config using the Vrf_address_family template
        vrf_address_family_parser = Vrf_address_familyTemplate(
            lines=finalConfig.splitlines(),
            module=self._module,
        )
        obj = vrf_address_family_parser.parse()
        objs = list(obj.values())

        for vrf in objs:
            af = vrf.get("address_families", {})
            if af:
                self._post_parse(vrf)
            else:
                vrf["address_families"] = []

        ansible_facts["ansible_network_resources"].pop("vrf_address_family", None)

        params = utils.remove_empties(
            vrf_address_family_parser.validate_config(
                self.argument_spec,
                {"config": objs},
                redact=True,
            ),
        )

        facts["vrf_address_family"] = params.get("config", [])
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts

    def _post_parse(self, af_data):
        """Converts the intermediate data structure
            to valid format as per argspec.
        :param obj: dict
        """
        af = af_data.get("address_families", {})
        if af:
            af_data["address_families"] = list(af.values())



# Copyright (c) 2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

"""Provide version object to compare version numbers."""

from __future__ import annotations

from ansible.module_utils.compat.version import LooseVersion

__all__ = ("LooseVersion",)
